# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.combinat sage.rings.finite_rings
r"""
Simplified DES

A simplified variant of the Data Encryption Standard (DES). Note that
Simplified DES or S-DES is for educational purposes only. It is a
small-scale version of the DES designed to help beginners understand the
basic structure of DES.

AUTHORS:

- Minh Van Nguyen (2009-06): initial version
"""

# #########################################################################
# Copyright (c) 2009 Minh Van Nguyen <nguyenminh2@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# https://www.gnu.org/licenses/
# #########################################################################

from sage.monoids.string_monoid import BinaryStrings
from sage.structure.sage_object import SageObject


class SimplifiedDES(SageObject):
    r"""
    This class implements the Simplified Data Encryption Standard (S-DES)
    described in [Sch1996]_. Schaefer's S-DES is for educational purposes
    only and is not secure for practical purposes. S-DES is a version of
    the DES with all parameters significantly reduced, but at the same time
    preserving the structure of DES. The goal of S-DES is to allow a
    beginner to understand the structure of DES, thus laying a foundation
    for a thorough study of DES. Its goal is as a teaching tool in the same
    spirit as Phan's
    :mod:`Mini-AES <sage.crypto.block_cipher.miniaes>` [Pha2002]_.

    EXAMPLES:

    Encrypt a random block of 8-bit plaintext using a random key, decrypt
    the ciphertext, and compare the result with the original plaintext::

        sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
        sage: sdes = SimplifiedDES(); sdes
        Simplified DES block cipher with 10-bit keys
        sage: bin = BinaryStrings()
        sage: P = [bin(str(randint(0, 1))) for i in range(8)]
        sage: K = sdes.random_key()
        sage: C = sdes.encrypt(P, K)
        sage: plaintxt = sdes.decrypt(C, K)
        sage: plaintxt == P
        True

    We can also encrypt binary strings that are larger than 8 bits in length.
    However, the number of bits in that binary string must be positive
    and a multiple of 8::

        sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
        sage: sdes = SimplifiedDES()
        sage: bin = BinaryStrings()
        sage: P = bin.encoding("Encrypt this using S-DES!")
        sage: Mod(len(P), 8) == 0
        True
        sage: K = sdes.list_to_string(sdes.random_key())
        sage: C = sdes(P, K, algorithm='encrypt')
        sage: plaintxt = sdes(C, K, algorithm='decrypt')
        sage: plaintxt == P
        True
    """

    def __init__(self):
        r"""
        A simplified variant of the Data Encryption Standard (DES).

        EXAMPLES::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES(); sdes
            Simplified DES block cipher with 10-bit keys
            sage: B = BinaryStrings()
            sage: P = [B(str(randint(0, 1))) for i in range(8)]
            sage: K = sdes.random_key()
            sage: C = sdes.encrypt(P, K)
            sage: plaintxt = sdes.decrypt(C, K)
            sage: plaintxt == P
            True
        """
        from sage.crypto.sbox import SBox
        # the number of bits in a secret key
        self._key_size = 10
        # the S-box S_0
        self._sbox0 = SBox(1, 0, 3, 2, 3, 2, 1, 0, 0, 2, 1, 3, 3, 1, 3, 2)
        # the S-box S_1
        self._sbox1 = SBox(0, 1, 2, 3, 2, 0, 1, 3, 3, 0, 1, 0, 2, 1, 0, 3)

    def __call__(self, B, K, algorithm='encrypt'):
        r"""
        Apply S-DES encryption or decryption on the binary string ``B``
        using the key ``K``.  The flag ``algorithm`` controls what action is
        to be performed on ``B``.

        INPUT:

        - ``B`` -- a binary string, where the number of bits is positive and
          a multiple of 8

        - ``K`` -- a secret key; this must be a 10-bit binary string

        - ``algorithm`` -- (default: ``'encrypt'``) a string; a flag to signify
          whether encryption or decryption is to be applied to the binary
          string ``B``. The encryption flag is ``'encrypt'`` and the decryption
          flag is ``'decrypt'``.

        OUTPUT:

        - The ciphertext (respectively plaintext) corresponding to the
          binary string ``B``.

        EXAMPLES:

        Encrypt a plaintext, decrypt the ciphertext, and compare the
        result with the original plaintext::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: bin = BinaryStrings()
            sage: P = bin.encoding("Encrypt this using DES!")
            sage: K = sdes.random_key()
            sage: K = sdes.list_to_string(K)
            sage: C = sdes(P, K, algorithm='encrypt')
            sage: plaintxt = sdes(C, K, algorithm='decrypt')
            sage: plaintxt == P
            True

        TESTS:

        The binary string ``B`` must be non-empty and the number of bits must
        be a multiple of 8::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes("B", "K")
            Traceback (most recent call last):
            ...
            TypeError: input B must be a non-empty binary string with number of bits a multiple of 8
            sage: bin = BinaryStrings()
            sage: B = bin("101")
            sage: sdes(B, "K")
            Traceback (most recent call last):
            ...
            ValueError: the number of bits in the binary string B must be positive and a multiple of 8

        The secret key ``K`` must be a block of 10 bits::

            sage: B = bin.encoding("abc")
            sage: sdes(B, "K")
            Traceback (most recent call last):
            ...
            TypeError: secret key must be a 10-bit binary string
            sage: K = bin("1010")
            sage: sdes(B, K)
            Traceback (most recent call last):
            ...
            ValueError: secret key must be a 10-bit binary string

        The value for ``algorithm`` must be either ``'encrypt'`` or
        ``'decrypt'``::

            sage: B = bin.encoding("abc")
            sage: K = sdes.list_to_string(sdes.random_key())
            sage: sdes(B, K, algorithm='e')
            Traceback (most recent call last):
            ...
            ValueError: algorithm must be either 'encrypt' or 'decrypt'
            sage: sdes(B, K, algorithm='d')
            Traceback (most recent call last):
            ...
            ValueError: algorithm must be either 'encrypt' or 'decrypt'
            sage: sdes(B, K, algorithm='abc')
            Traceback (most recent call last):
            ...
            ValueError: algorithm must be either 'encrypt' or 'decrypt'
        """
        from sage.monoids.string_monoid_element import StringMonoidElement
        from sage.rings.finite_rings.integer_mod import Mod
        # S-DES operates on 8-bit ciphertext/plaintext blocks
        Blength = 8

        if not isinstance(B, StringMonoidElement):
            raise TypeError("input B must be a non-empty binary string with number of bits a multiple of 8")
        if (len(B) == 0) or (Mod(len(B), Blength).lift() != 0):
            raise ValueError("the number of bits in the binary string B must be positive and a multiple of 8")
        if not isinstance(K, StringMonoidElement):
            raise TypeError("secret key must be a 10-bit binary string")
        if len(K) != self._key_size:
            raise ValueError("secret key must be a 10-bit binary string")

        N = len(B) // Blength  # the number of 8-bit blocks
        S = ""
        bin = BinaryStrings()
        # encrypt each 8-bit block in succession
        if algorithm == "encrypt":
            for i in range(N):
                # get an 8-bit block
                block = B[i*Blength : (i+1)*Blength]
                block = self.string_to_list(str(block))
                key = self.string_to_list(str(K))
                # encrypt the block using key
                C = self.encrypt(block, key)
                C = self.list_to_string(C)
                # append encrypted block to ciphertext string
                S = "".join([S, str(C)])
            return bin(S)
        # decrypt each 8-bit block in succession
        elif algorithm == "decrypt":
            for i in range(N):
                # get an 8-bit block
                block = B[i*Blength : (i+1)*Blength]
                block = self.string_to_list(str(block))
                key = self.string_to_list(str(K))
                # decrypt the block using key
                P = self.decrypt(block, key)
                P = self.list_to_string(P)
                # append decrypted block to plaintext string
                S = "".join([S, str(P)])
            return bin(S)
        # invalid value for algorithm option
        else:
            raise ValueError("algorithm must be either 'encrypt' or 'decrypt'")

    def __eq__(self, other):
        r"""
        Compare ``self`` with ``other``.

        Simplified DES objects are the same if they have the same key size
        and S-boxes.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: s = SimplifiedDES()
            sage: s == loads(dumps(s))
            True
        """
        return ( (self._key_size == other._key_size) and
                 (self._sbox0 == other._sbox0) and
                 (self._sbox1 == other._sbox1) )

    def __repr__(self):
        r"""
        A string representation of this Simplified DES.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: SimplifiedDES()
            Simplified DES block cipher with 10-bit keys
        """
        return "Simplified DES block cipher with 10-bit keys"

    def block_length(self):
        r"""
        Return the block length of Schaefer's S-DES block cipher. A key in
        Schaefer's S-DES is a block of 10 bits.

        OUTPUT:

        - The block (or key) length in number of bits.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.block_length()
            10
        """
        return self._key_size

    def decrypt(self, C, K):
        r"""
        Return an 8-bit plaintext corresponding to the ciphertext ``C``,
        using S-DES decryption with key ``K``. The decryption process of
        S-DES is as follows. Let `P` be the initial permutation function,
        `P^{-1}` the corresponding inverse permutation, `\Pi_F` the
        permutation/substitution function, and `\sigma` the switch function.
        The ciphertext block ``C`` first goes through `P`, the output of
        which goes through `\Pi_F` using the second subkey. Then we apply
        the switch function to the output of the last function, and the
        result is then fed into `\Pi_F` using the first subkey. Finally,
        run the output through `P^{-1}` to get the plaintext.

        INPUT:

        - ``C`` -- an 8-bit ciphertext; a block of 8 bits

        - ``K`` -- a 10-bit key; a block of 10 bits

        OUTPUT:

        The 8-bit plaintext corresponding to ``C``, obtained using the
        key ``K``.

        EXAMPLES:

        Decrypt an 8-bit ciphertext block::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: C = [0, 1, 0, 1, 0, 1, 0, 1]
            sage: K = [1, 0, 1, 0, 0, 0, 0, 0, 1, 0]
            sage: sdes.decrypt(C, K)
            [0, 0, 0, 1, 0, 1, 0, 1]

        We can also work with strings of bits::

            sage: C = "01010101"
            sage: K = "1010000010"
            sage: sdes.decrypt(sdes.string_to_list(C), sdes.string_to_list(K))
            [0, 0, 0, 1, 0, 1, 0, 1]

        TESTS:

        The ciphertext must be a block of 8 bits::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.decrypt("C", "K")
            Traceback (most recent call last):
            ...
            TypeError: ciphertext must be a list of 8 bits
            sage: sdes.decrypt([], "K")
            Traceback (most recent call last):
            ...
            ValueError: ciphertext must be a list of 8 bits
            sage: sdes.decrypt([1, 2, 3, 4], "K")
            Traceback (most recent call last):
            ...
            ValueError: ciphertext must be a list of 8 bits

        The key must be a block of 10 bits::

            sage: sdes.decrypt([1, 0, 1, 0, 1, 1, 0, 1], "K")
            Traceback (most recent call last):
            ...
            TypeError: the key must be a list of 10 bits
            sage: sdes.decrypt([1, 0, 1, 0, 1, 1, 0, 1], [])
            Traceback (most recent call last):
            ...
            TypeError: the key must be a list of 10 bits
            sage: sdes.decrypt([1, 0, 1, 0, 1, 1, 0, 1], [1, 2, 3, 4, 5])
            Traceback (most recent call last):
            ...
            TypeError: the key must be a list of 10 bits

        The value of each element of ``C`` or ``K`` must be either 0 or 1::

            sage: C = [1, 2, 3, 4, 5, 6, 7, 8]
            sage: K = [11, 12, 13, 14, 15, 16, 17, 18, 19, 20]
            sage: sdes.decrypt(C, K)
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 2) is not a valid string.
            sage: C = [0, 1, 0, 0, 1, 1, 1, 0]
            sage: K = [11, 12, 13, 14, 15, 16, 17, 18, 19, 20]
            sage: sdes.decrypt(C, K)
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 13) is not a valid string.
        """
        # sanity check
        if not isinstance(C, list):
            raise TypeError("ciphertext must be a list of 8 bits")
        if len(C) != 8:
            raise ValueError("ciphertext must be a list of 8 bits")
        if not isinstance(K, list):
            raise TypeError("the key must be a list of 10 bits")
        if len(K) != 10:
            raise TypeError("the key must be a list of 10 bits")

        # run through initial permutation
        P = self.initial_permutation(C, inverse=False)
        # run through Pi_F with subkey 2
        P = self.permute_substitute(P, self.subkey(K, n=2))
        # run through switch function
        P = self.switch(P)
        # run through Pi_F with subkey 1
        P = self.permute_substitute(P, self.subkey(K, n=1))
        # run through inverse permutation
        P = self.initial_permutation(P, inverse=True)
        # output the plaintext
        return P

    def encrypt(self, P, K):
        r"""
        Return an 8-bit ciphertext corresponding to the plaintext ``P``,
        using S-DES encryption with key ``K``. The encryption process of
        S-DES is as follows. Let `P` be the initial permutation function,
        `P^{-1}` the corresponding inverse permutation, `\Pi_F` the
        permutation/substitution function, and `\sigma` the switch function.
        The plaintext block ``P`` first goes through `P`, the output of
        which goes through `\Pi_F` using the first subkey. Then we apply
        the switch function to the output of the last function, and the
        result is then fed into `\Pi_F` using the second subkey. Finally,
        run the output through `P^{-1}` to get the ciphertext.

        INPUT:

        - ``P`` -- an 8-bit plaintext; a block of 8 bits

        - ``K`` -- a 10-bit key; a block of 10 bits

        OUTPUT:

        The 8-bit ciphertext corresponding to ``P``, obtained using the
        key ``K``.

        EXAMPLES:

        Encrypt an 8-bit plaintext block::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: P = [0, 1, 0, 1, 0, 1, 0, 1]
            sage: K = [1, 0, 1, 0, 0, 0, 0, 0, 1, 0]
            sage: sdes.encrypt(P, K)
            [1, 1, 0, 0, 0, 0, 0, 1]

        We can also work with strings of bits::

            sage: P = "01010101"
            sage: K = "1010000010"
            sage: sdes.encrypt(sdes.string_to_list(P), sdes.string_to_list(K))
            [1, 1, 0, 0, 0, 0, 0, 1]

        TESTS:

        The plaintext must be a block of 8 bits::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.encrypt("P", "K")
            Traceback (most recent call last):
            ...
            TypeError: plaintext must be a list of 8 bits
            sage: sdes.encrypt([], "K")
            Traceback (most recent call last):
            ...
            ValueError: plaintext must be a list of 8 bits
            sage: sdes.encrypt([1, 2, 3, 4], "K")
            Traceback (most recent call last):
            ...
            ValueError: plaintext must be a list of 8 bits

        The key must be a block of 10 bits::

            sage: sdes.encrypt([1, 0, 1, 0, 1, 1, 0, 1], "K")
            Traceback (most recent call last):
            ...
            TypeError: the key must be a list of 10 bits
            sage: sdes.encrypt([1, 0, 1, 0, 1, 1, 0, 1], [])
            Traceback (most recent call last):
            ...
            TypeError: the key must be a list of 10 bits
            sage: sdes.encrypt([1, 0, 1, 0, 1, 1, 0, 1], [1, 2, 3, 4, 5])
            Traceback (most recent call last):
            ...
            TypeError: the key must be a list of 10 bits

        The value of each element of ``P`` or ``K`` must be either 0 or 1::

            sage: P = [1, 2, 3, 4, 5, 6, 7, 8]
            sage: K = [11, 12, 13, 14, 15, 16, 17, 18, 19, 20]
            sage: sdes.encrypt(P, K)
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 2) is not a valid string.
            sage: P = [0, 1, 0, 0, 1, 1, 1, 0]
            sage: K = [11, 12, 13, 14, 15, 16, 17, 18, 19, 20]
            sage: sdes.encrypt(P, K)
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 13) is not a valid string.
        """
        # sanity check
        if not isinstance(P, list):
            raise TypeError("plaintext must be a list of 8 bits")
        if len(P) != 8:
            raise ValueError("plaintext must be a list of 8 bits")
        if not isinstance(K, list):
            raise TypeError("the key must be a list of 10 bits")
        if len(K) != 10:
            raise TypeError("the key must be a list of 10 bits")

        # run through initial permutation
        C = self.initial_permutation(P, inverse=False)
        # run through Pi_F with subkey 1
        C = self.permute_substitute(C, self.subkey(K, n=1))
        # run through switch function
        C = self.switch(C)
        # run through Pi_F with subkey 2
        C = self.permute_substitute(C, self.subkey(K, n=2))
        # run through inverse permutation
        C = self.initial_permutation(C, inverse=True)
        # output the ciphertext
        return C

    def initial_permutation(self, B, inverse=False):
        r"""
        Return the initial permutation of ``B``. Denote the initial
        permutation function by `P` and let `(b_0, b_1, b_2, \dots, b_7)`
        be a vector of 8 bits, where each `b_i \in \{ 0, 1 \}`. Then

        .. MATH::


            P(b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7)
            = (b_1, b_5, b_2, b_0, b_3, b_7, b_4, b_6)

        The inverse permutation is `P^{-1}`:

        .. MATH::

            P^{-1}(b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7)
            = (b_3, b_0, b_2, b_4, b_6, b_1, b_7, b_5)

        INPUT:

        - ``B`` -- list; a block of 8 bits

        - ``inverse`` -- boolean (default: ``False``); if ``True`` then use the
          inverse permutation `P^{-1}`. If ``False`` then use the initial
          permutation `P`.

        OUTPUT:

        The initial permutation of ``B`` if ``inverse=False``, or the
        inverse permutation of ``B`` if ``inverse=True``.

        EXAMPLES:

        The initial permutation of a list of 8 bits::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: B = [1, 0, 1, 1, 0, 1, 0, 0]
            sage: P = sdes.initial_permutation(B); P
            [0, 1, 1, 1, 1, 0, 0, 0]

        Recovering the original list of 8 bits from the permutation::

            sage: Pinv = sdes.initial_permutation(P, inverse=True)
            sage: Pinv; B
            [1, 0, 1, 1, 0, 1, 0, 0]
            [1, 0, 1, 1, 0, 1, 0, 0]

        We can also work with a string of bits::

            sage: S = "10110100"
            sage: L = sdes.string_to_list(S)
            sage: P = sdes.initial_permutation(L); P
            [0, 1, 1, 1, 1, 0, 0, 0]
            sage: sdes.initial_permutation(sdes.string_to_list("01111000"), inverse=True)
            [1, 0, 1, 1, 0, 1, 0, 0]

        TESTS:

        The input block must be a list::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.initial_permutation("B")
            Traceback (most recent call last):
            ...
            TypeError: input block must be a list of 8 bits
            sage: sdes.initial_permutation(())
            Traceback (most recent call last):
            ...
            TypeError: input block must be a list of 8 bits

        The input block must be a list of 8 bits::

            sage: sdes.initial_permutation([])
            Traceback (most recent call last):
            ...
            ValueError: input block must be a list of 8 bits
            sage: sdes.initial_permutation([1, 2, 3, 4, 5, 6, 7, 8, 9])
            Traceback (most recent call last):
            ...
            ValueError: input block must be a list of 8 bits

        The value of each element of the list must be either 0 or 1::

            sage: sdes.initial_permutation([1, 2, 3, 4, 5, 6, 7, 8])
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 2) is not a valid string.
        """
        # sanity check
        if not isinstance(B, list):
            raise TypeError("input block must be a list of 8 bits")
        if len(B) != 8:
            raise ValueError("input block must be a list of 8 bits")

        bin = BinaryStrings()

        # use the initial permutation P
        if not inverse:
            return [ bin(str(B[1])), bin(str(B[5])),
                     bin(str(B[2])), bin(str(B[0])),
                     bin(str(B[3])), bin(str(B[7])),
                     bin(str(B[4])), bin(str(B[6])) ]

        # use the inverse permutation P^-1
        if inverse:
            return [ bin(str(B[3])), bin(str(B[0])),
                     bin(str(B[2])), bin(str(B[4])),
                     bin(str(B[6])), bin(str(B[1])),
                     bin(str(B[7])), bin(str(B[5])) ]

    def left_shift(self, B, n=1):
        r"""
        Return a circular left shift of ``B`` by ``n`` positions. Let
        `B = (b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7, b_8, b_9)` be a vector
        of 10 bits. Then the left shift operation `L_n` is performed on the
        first 5 bits and the last 5 bits of `B` separately. That is, if the
        number of shift positions is ``n=1``, then `L_1` is defined as

        .. MATH::

            L_1(b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7, b_8, b_9)
            = (b_1, b_2, b_3, b_4, b_0, b_6, b_7, b_8, b_9, b_5)

        If the number of shift positions is ``n=2``, then `L_2` is given by

        .. MATH::

            L_2(b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7, b_8, b_9)
            = (b_2, b_3, b_4, b_0, b_1, b_7, b_8, b_9, b_5, b_6)

        INPUT:

        - ``B`` -- list of 10 bits

        - ``n`` -- (default: 1) if ``n=1`` then perform left shift by 1
          position; if ``n=2`` then perform left shift by 2 positions. The
          valid values for ``n`` are 1 and 2, since only up to 2 positions
          are defined for this circular left shift operation.

        OUTPUT: the circular left shift of each half of ``B``

        EXAMPLES:

        Circular left shift by 1 position of a 10-bit string::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: B = [1, 0, 0, 0, 0, 0, 1, 1, 0, 0]
            sage: sdes.left_shift(B)
            [0, 0, 0, 0, 1, 1, 1, 0, 0, 0]
            sage: sdes.left_shift([1, 0, 1, 0, 0, 0, 0, 0, 1, 0])
            [0, 1, 0, 0, 1, 0, 0, 1, 0, 0]

        Circular left shift by 2 positions of a 10-bit string::

            sage: B = [0, 0, 0, 0, 1, 1, 1, 0, 0, 0]
            sage: sdes.left_shift(B, n=2)
            [0, 0, 1, 0, 0, 0, 0, 0, 1, 1]

        Here we work with a string of bits::

            sage: S = "1000001100"
            sage: L = sdes.string_to_list(S)
            sage: sdes.left_shift(L)
            [0, 0, 0, 0, 1, 1, 1, 0, 0, 0]
            sage: sdes.left_shift(sdes.string_to_list("1010000010"), n=2)
            [1, 0, 0, 1, 0, 0, 1, 0, 0, 0]

        TESTS:

        The input block must be a list::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.left_shift("B")
            Traceback (most recent call last):
            ...
            TypeError: input block must be a list of 10 bits
            sage: sdes.left_shift(())
            Traceback (most recent call last):
            ...
            TypeError: input block must be a list of 10 bits

        The input block must be a list of 10 bits::

            sage: sdes.left_shift([])
            Traceback (most recent call last):
            ...
            ValueError: input block must be a list of 10 bits
            sage: sdes.left_shift([1, 2, 3, 4, 5])
            Traceback (most recent call last):
            ...
            ValueError: input block must be a list of 10 bits

        The value of each element of the list must be either 0 or 1::

            sage: sdes.left_shift([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 2) is not a valid string.

        The number of shift positions must be either 1 or 2::

            sage: B = [0, 0, 0, 0, 1, 1, 1, 0, 0, 0]
            sage: sdes.left_shift(B, n=-1)
            Traceback (most recent call last):
            ...
            ValueError: input n must be either 1 or 2
            sage: sdes.left_shift(B, n=3)
            Traceback (most recent call last):
            ...
            ValueError: input n must be either 1 or 2
        """
        # sanity check
        if not isinstance(B, list):
            raise TypeError("input block must be a list of 10 bits")
        if len(B) != 10:
            raise ValueError("input block must be a list of 10 bits")

        bin = BinaryStrings()
        # circular left shift by 1 position
        if n == 1:
            return [ bin(str(B[1])), bin(str(B[2])),
                     bin(str(B[3])), bin(str(B[4])),
                     bin(str(B[0])), bin(str(B[6])),
                     bin(str(B[7])), bin(str(B[8])),
                     bin(str(B[9])), bin(str(B[5])) ]
        # circular left shift by 2 positions
        elif n == 2:
            return [ bin(str(B[2])), bin(str(B[3])),
                     bin(str(B[4])), bin(str(B[0])),
                     bin(str(B[1])), bin(str(B[7])),
                     bin(str(B[8])), bin(str(B[9])),
                     bin(str(B[5])), bin(str(B[6])) ]
        # an invalid number of shift positions
        else:
            raise ValueError("input n must be either 1 or 2")

    def list_to_string(self, B):
        r"""
        Return a binary string representation of the list ``B``.

        INPUT:

        - ``B`` -- a non-empty list of bits

        OUTPUT: the binary string representation of ``B``

        EXAMPLES:

        A binary string representation of a list of bits::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: L = [0, 0, 0, 0, 1, 1, 0, 1, 0, 0]
            sage: sdes.list_to_string(L)
            0000110100

        TESTS:

        Input ``B`` must be a non-empty list::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.list_to_string("L")
            Traceback (most recent call last):
            ...
            TypeError: input B must be a non-empty list of bits
            sage: sdes.list_to_string([])
            Traceback (most recent call last):
            ...
            ValueError: input B must be a non-empty list of bits

        Input must be a non-empty list of bits::

            sage: sdes.list_to_string([0, 1, 2])
            <repr(<sage.monoids.string_monoid_element.StringMonoidElement at 0x...>) failed: IndexError: tuple index out of range>
        """
        # sanity check
        if not isinstance(B, list):
            raise TypeError("input B must be a non-empty list of bits")
        if len(B) == 0:
            raise ValueError("input B must be a non-empty list of bits")

        # perform the conversion from list to binary string
        from sage.rings.integer import Integer
        bin = BinaryStrings()
        return bin([Integer(str(b)) for b in B])

    def permutation4(self, B):
        r"""
        Return a permutation of a 4-bit string. This permutation is called
        `P_4` and is specified as follows. Let
        `(b_0, b_1, b_2, b_3)` be a vector of 4 bits where each
        `b_i \in \{ 0, 1 \}`. Then `P_4` is defined by

        .. MATH::

            P_4(b_0, b_1, b_2, b_3) = (b_1, b_3, b_2, b_0)

        INPUT:

        - ``B`` -- a block of 4-bit string

        OUTPUT: a permutation of ``B``

        EXAMPLES:

        Permute a 4-bit string::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: B = [1, 1, 0, 0]
            sage: sdes.permutation4(B)
            [1, 0, 0, 1]
            sage: sdes.permutation4([0, 1, 0, 1])
            [1, 1, 0, 0]

        We can also work with a string of bits::

            sage: S = "1100"
            sage: L = sdes.string_to_list(S)
            sage: sdes.permutation4(L)
            [1, 0, 0, 1]
            sage: sdes.permutation4(sdes.string_to_list("0101"))
            [1, 1, 0, 0]

        TESTS:

        The input block must be a list::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.permutation4("B")
            Traceback (most recent call last):
            ...
            TypeError: input block must be a list of 4 bits
            sage: sdes.permutation4(())
            Traceback (most recent call last):
            ...
            TypeError: input block must be a list of 4 bits

        The input block must be a list of 4 bits::

            sage: sdes.permutation4([])
            Traceback (most recent call last):
            ...
            ValueError: input block must be a list of 4 bits
            sage: sdes.permutation4([1, 2, 3, 4, 5])
            Traceback (most recent call last):
            ...
            ValueError: input block must be a list of 4 bits

        The value of each element of the list must be either 0 or 1::

            sage: sdes.permutation4([1, 2, 3, 4])
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 2) is not a valid string.
        """
        # sanity check
        if not isinstance(B, list):
            raise TypeError("input block must be a list of 4 bits")
        if len(B) != 4:
            raise ValueError("input block must be a list of 4 bits")

        # perform the permutation
        bin = BinaryStrings()
        return [ bin(str(B[1])), bin(str(B[3])),
                 bin(str(B[2])), bin(str(B[0])) ]

    def permutation8(self, B):
        r"""
        Return a permutation of an 8-bit string. This permutation is called
        `P_8` and is specified as follows. Let
        `(b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7, b_8, b_9)` be a vector of
        10 bits where each `b_i \in \{ 0, 1 \}`. Then `P_8` picks out 8 of
        those 10 bits and permutes those 8 bits:

        .. MATH::

            P_8(b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7, b_8, b_9)
            =
            (b_5, b_2, b_6, b_3, b_7, b_4, b_9, b_8)

        INPUT:

        - ``B`` -- a block of 10-bit string

        OUTPUT: pick out 8 of the 10 bits of ``B`` and permute those 8 bits

        EXAMPLES:

        Permute a 10-bit string::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: B = [1, 1, 0, 0, 1, 0, 0, 1, 0, 1]
            sage: sdes.permutation8(B)
            [0, 0, 0, 0, 1, 1, 1, 0]
            sage: sdes.permutation8([0, 1, 1, 0, 1, 0, 0, 1, 0, 1])
            [0, 1, 0, 0, 1, 1, 1, 0]
            sage: sdes.permutation8([0, 0, 0, 0, 1, 1, 1, 0, 0, 0])
            [1, 0, 1, 0, 0, 1, 0, 0]

        We can also work with a string of bits::

            sage: S = "1100100101"
            sage: L = sdes.string_to_list(S)
            sage: sdes.permutation8(L)
            [0, 0, 0, 0, 1, 1, 1, 0]
            sage: sdes.permutation8(sdes.string_to_list("0110100101"))
            [0, 1, 0, 0, 1, 1, 1, 0]

        TESTS:

        The input block must be a list::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.permutation8("B")
            Traceback (most recent call last):
            ...
            TypeError: input block must be a list of 10 bits
            sage: sdes.permutation8(())
            Traceback (most recent call last):
            ...
            TypeError: input block must be a list of 10 bits

        The input block must be a list of 10 bits::

            sage: sdes.permutation8([])
            Traceback (most recent call last):
            ...
            ValueError: input block must be a list of 10 bits
            sage: sdes.permutation8([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
            Traceback (most recent call last):
            ...
            ValueError: input block must be a list of 10 bits

        The value of each element of the list must be either 0 or 1::

            sage: sdes.permutation8([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 6) is not a valid string.
        """
        # sanity check
        if not isinstance(B, list):
            raise TypeError("input block must be a list of 10 bits")
        if len(B) != 10:
            raise ValueError("input block must be a list of 10 bits")

        # perform the permutation
        bin = BinaryStrings()
        return [ bin(str(B[5])), bin(str(B[2])),
                 bin(str(B[6])), bin(str(B[3])),
                 bin(str(B[7])), bin(str(B[4])),
                 bin(str(B[9])), bin(str(B[8])) ]

    def permutation10(self, B):
        r"""
        Return a permutation of a 10-bit string. This permutation is called
        `P_{10}` and is specified as follows. Let
        `(b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7, b_8, b_9)` be a vector of
        10 bits where each `b_i \in \{ 0, 1 \}`. Then `P_{10}` is given by

        .. MATH::

            P_{10}(b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7, b_8, b_9)
            =
            (b_2, b_4, b_1, b_6, b_3, b_9, b_0, b_8, b_7, b_5)

        INPUT:

        - ``B`` -- a block of 10-bit string

        OUTPUT: a permutation of ``B``

        EXAMPLES:

        Permute a 10-bit string::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: B = [1, 1, 0, 0, 1, 0, 0, 1, 0, 1]
            sage: sdes.permutation10(B)
            [0, 1, 1, 0, 0, 1, 1, 0, 1, 0]
            sage: sdes.permutation10([0, 1, 1, 0, 1, 0, 0, 1, 0, 1])
            [1, 1, 1, 0, 0, 1, 0, 0, 1, 0]
            sage: sdes.permutation10([1, 0, 1, 0, 0, 0, 0, 0, 1, 0])
            [1, 0, 0, 0, 0, 0, 1, 1, 0, 0]

        Here we work with a string of bits::

            sage: S = "1100100101"
            sage: L = sdes.string_to_list(S)
            sage: sdes.permutation10(L)
            [0, 1, 1, 0, 0, 1, 1, 0, 1, 0]
            sage: sdes.permutation10(sdes.string_to_list("0110100101"))
            [1, 1, 1, 0, 0, 1, 0, 0, 1, 0]

        TESTS:

        The input block must be a list::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.permutation10("B")
            Traceback (most recent call last):
            ...
            TypeError: input block must be a list of 10 bits
            sage: sdes.permutation10(())
            Traceback (most recent call last):
            ...
            TypeError: input block must be a list of 10 bits

        The input block must be a list of 10 bits::

            sage: sdes.permutation10([])
            Traceback (most recent call last):
            ...
            ValueError: input block must be a list of 10 bits
            sage: sdes.permutation10([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
            Traceback (most recent call last):
            ...
            ValueError: input block must be a list of 10 bits

        The value of each element of the list must be either 0 or 1::

            sage: sdes.permutation10([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 3) is not a valid string.
        """
        # sanity check
        if not isinstance(B, list):
            raise TypeError("input block must be a list of 10 bits")
        if len(B) != 10:
            raise ValueError("input block must be a list of 10 bits")

        # perform the permutation
        bin = BinaryStrings()
        return [ bin(str(B[2])), bin(str(B[4])),
                 bin(str(B[1])), bin(str(B[6])),
                 bin(str(B[3])), bin(str(B[9])),
                 bin(str(B[0])), bin(str(B[8])),
                 bin(str(B[7])), bin(str(B[5])) ]

    def permute_substitute(self, B, key):
        r"""
        Apply the function `\Pi_F` on the block ``B`` using subkey ``key``.
        Let `(b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7)`
        be a vector of 8 bits where each `b_i \in \{ 0, 1 \}`, let `L` and
        `R` be the leftmost 4 bits and rightmost 4 bits of ``B``
        respectively, and let `F` be a function mapping 4-bit strings to
        4-bit strings. Then

        .. MATH::

            \Pi_F(L, R) = (L \oplus F(R, S), R)

        where `S` is a subkey and `\oplus` denotes the bit-wise
        exclusive-OR function.

        The function `F` can be described as follows. Its 4-bit input block
        `(n_0, n_1, n_2, n_3)` is first expanded into an 8-bit block
        to become `(n_3, n_0, n_1, n_2, n_1, n_2, n_3, n_0)`. This is
        usually represented as follows

        .. MATH::

            \begin{tabular}{c|cc|c}
              $n_3$ & $n_0$ & $n_1$ & $n_2$ \\
              $n_1$ & $n_2$ & $n_3$ & $n_0$
            \end{tabular}

        Let `K = (k_0, k_1, k_2, k_3, k_4, k_5, k_6, k_7)` be an 8-bit
        subkey. Then `K` is added to the above expanded input block using
        exclusive-OR to produce

        .. MATH::

            \begin{tabular}{c|cc|c}
              $n_3 + k_0$ & $n_0 + k_1$ & $n_1 + k_2$ & $n_2 + k_3$ \\
              $n_1 + k_4$ & $n_2 + k_5$ & $n_3 + k_6$ & $n_0 + k_7$
            \end{tabular}
            =
            \begin{tabular}{c|cc|c}
              $p_{0,0}$ & $p_{0,1}$ & $p_{0,2}$ & $p_{0,3}$ \\
              $p_{1,0}$ & $p_{1,1}$ & $p_{1,2}$ & $p_{1,3}$
            \end{tabular}

        Now read the first row as the 4-bit string
        `p_{0,0} p_{0,3} p_{0,1} p_{0,2}` and input this 4-bit string through
        S-box `S_0` to get a 2-bit output.

        .. MATH::

            S_0
            =
            \begin{tabular}{cc|cc}            \hline
              Input & Output & Input & Output \\\hline
              0000  & 01     & 1000  & 00     \\
              0001  & 00     & 1001  & 10     \\
              0010  & 11     & 1010  & 01     \\
              0011  & 10     & 1011  & 11     \\
              0100  & 11     & 1100  & 11     \\
              0101  & 10     & 1101  & 01     \\
              0110  & 01     & 1110  & 11     \\
              0111  & 00     & 1111  & 10     \\\hline
            \end{tabular}

        Next read the second row as the 4-bit string
        `p_{1,0} p_{1,3} p_{1,1} p_{1,2}` and input this 4-bit string through
        S-box `S_1` to get another 2-bit output.

        .. MATH::

            S_1
            =
            \begin{tabular}{cc|cc}            \hline
              Input & Output & Input & Output \\\hline
              0000  & 00     & 1000  & 11     \\
              0001  & 01     & 1001  & 00     \\
              0010  & 10     & 1010  & 01     \\
              0011  & 11     & 1011  & 00     \\
              0100  & 10     & 1100  & 10     \\
              0101  & 00     & 1101  & 01     \\
              0110  & 01     & 1110  & 00     \\
              0111  & 11     & 1111  & 11     \\\hline
            \end{tabular}

        Denote the 4 bits produced by `S_0` and `S_1` as `b_0 b_1 b_2 b_3`.
        This 4-bit string undergoes another permutation called `P_4` as
        follows:

        .. MATH::

            P_4(b_0, b_1, b_2, b_3) = (b_1, b_3, b_2, b_0)

        The output of `P_4` is the output of the function `F`.

        INPUT:

        - ``B`` -- list of 8 bits

        - ``key`` -- an 8-bit subkey

        OUTPUT: the result of applying the function `\Pi_F` to ``B``

        EXAMPLES:

        Applying the function `\Pi_F` to an 8-bit block and an 8-bit subkey::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: B = [1, 0, 1, 1, 1, 1, 0, 1]
            sage: K = [1, 1, 0, 1, 0, 1, 0, 1]
            sage: sdes.permute_substitute(B, K)
            [1, 0, 1, 0, 1, 1, 0, 1]

        We can also work with strings of bits::

            sage: B = "10111101"
            sage: K = "11010101"
            sage: B = sdes.string_to_list(B); K = sdes.string_to_list(K)
            sage: sdes.permute_substitute(B, K)
            [1, 0, 1, 0, 1, 1, 0, 1]

        TESTS:

        The input ``B`` must be a block of 8 bits::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.permute_substitute("B", "K")
            Traceback (most recent call last):
            ...
            TypeError: input B must be an 8-bit string
            sage: sdes.permute_substitute([], "K")
            Traceback (most recent call last):
            ...
            ValueError: input B must be an 8-bit string

        The input ``key`` must be an 8-bit subkey::

            sage: sdes.permute_substitute([0, 1, 0, 0, 1, 1, 1, 0], "K")
            Traceback (most recent call last):
            ...
            TypeError: input key must be an 8-bit subkey
            sage: sdes.permute_substitute([0, 1, 0, 0, 1, 1, 1, 0], [])
            Traceback (most recent call last):
            ...
            ValueError: input key must be an 8-bit subkey

        The value of each element of ``B`` or ``key`` must be either 0 or 1::

            sage: B = [1, 2, 3, 4, 5, 6, 7, 8]
            sage: K = [0, 1, 2, 3, 4, 5, 6, 7]
            sage: sdes.permute_substitute(B, K)
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 2) is not a valid string.
            sage: B = [0, 1, 0, 0, 1, 1, 1, 0]
            sage: K = [1, 2, 3, 4, 5, 6, 7, 8]
            sage: sdes.permute_substitute(B, K)
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 2) is not a valid string.
        """
        # sanity check
        if not isinstance(B, list):
            raise TypeError("input B must be an 8-bit string")
        if len(B) != 8:
            raise ValueError("input B must be an 8-bit string")
        if not isinstance(key, list):
            raise TypeError("input key must be an 8-bit subkey")
        if len(key) != 8:
            raise ValueError("input key must be an 8-bit subkey")

        from sage.rings.finite_rings.finite_field_constructor import FiniteField
        GF = FiniteField(2, "x")
        bin = BinaryStrings()
        bin_to_GF2 = {bin("0"): GF(0), bin("1"): GF(1)}

        # the leftmost 4 bits of B
        L = [ bin_to_GF2[bin(str(B[i]))] for i in range(4) ]
        # the rightmost 4 bits of B
        R = [ bin_to_GF2[bin(str(B[i]))] for i in range(4, len(B)) ]
        # get the GF(2) representation of the subkey
        K = [ bin_to_GF2[bin(str(key[i]))] for i in range(len(key)) ]
        # expand the rightmost 4 bits into an 8-bit block
        RX = [ R[3], R[0], R[1], R[2], R[1], R[2], R[3], R[0] ]
        # add the subkey to the expanded 8-bit block using exclusive-OR
        P = [ RX[i] + K[i] for i in range(len(K)) ]
        # run each half of P separately through the S-boxes
        left = self._sbox0([ P[0], P[3], P[1], P[2] ])
        right = self._sbox1([ P[4], P[7], P[5], P[6] ])
        # First concatenate the left and right parts, then get the
        # output of the function F.
        F = self.permutation4(left + right)
        F = [ bin_to_GF2[F[i]] for i in range(len(F)) ]
        # Add L to F using exclusive-OR. Then concatenate the result with
        # the rightmost 4 bits of B. This is the output of the function Pi_F.
        L = [ L[i] + F[i] for i in range(len(F)) ]
        return L + R

    def random_key(self):
        r"""
        Return a random 10-bit key.

        EXAMPLES:

        The size of each key is the same as the block size::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: key = sdes.random_key()
            sage: len(key) == sdes.block_length()
            True
        """
        from sage.misc.prandom import randint
        bin = BinaryStrings()
        return [bin(str(randint(0, 1))) for i in range(self._key_size)]

    def sbox(self):
        r"""
        Return the S-boxes of simplified DES.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sbox = sdes.sbox()
            sage: sbox[0]; sbox[1]
            (1, 0, 3, 2, 3, 2, 1, 0, 0, 2, 1, 3, 3, 1, 3, 2)
            (0, 1, 2, 3, 2, 0, 1, 3, 3, 0, 1, 0, 2, 1, 0, 3)
        """
        return [self._sbox0, self._sbox1]

    def string_to_list(self, S):
        r"""
        Return a list representation of the binary string ``S``.

        INPUT:

        - ``S`` -- string of bits

        OUTPUT: list representation of the string ``S``

        EXAMPLES:

        A list representation of a string of bits::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: S = "0101010110"
            sage: sdes.string_to_list(S)
            [0, 1, 0, 1, 0, 1, 0, 1, 1, 0]

        TESTS:

        Input must be a non-empty string::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.string_to_list("")
            Traceback (most recent call last):
            ...
            ValueError: input S must be a non-empty string of bits
            sage: sdes.string_to_list(1)
            Traceback (most recent call last):
            ...
            TypeError: input S must be a non-empty string of bits

        Input must be a non-empty string of bits::

            sage: sdes.string_to_list("0123")
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 2) is not a valid string.
        """
        # sanity check
        if not isinstance(S, str):
            raise TypeError("input S must be a non-empty string of bits")
        if len(S) == 0:
            raise ValueError("input S must be a non-empty string of bits")

        # perform the conversion from string to list
        bin = BinaryStrings()
        return [bin(s) for s in S]

    def subkey(self, K, n=1):
        r"""
        Return the `n`-th subkey based on the key ``K``.

        INPUT:

        - ``K`` -- a 10-bit secret key of this Simplified DES

        - ``n`` -- (default: 1) if ``n=1`` then return the first subkey based
          on ``K``; if ``n=2`` then return the second subkey. The valid
          values for ``n`` are 1 and 2, since only two subkeys are defined
          for each secret key in Schaefer's S-DES.

        OUTPUT: the `n`-th subkey based on the secret key ``K``

        EXAMPLES:

        Obtain the first subkey from a secret key::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: key = [1, 0, 1, 0, 0, 0, 0, 0, 1, 0]
            sage: sdes.subkey(key, n=1)
            [1, 0, 1, 0, 0, 1, 0, 0]

        Obtain the second subkey from a secret key::

            sage: key = [1, 0, 1, 0, 0, 0, 0, 0, 1, 0]
            sage: sdes.subkey(key, n=2)
            [0, 1, 0, 0, 0, 0, 1, 1]

        We can also work with strings of bits::

            sage: K = "1010010010"
            sage: L = sdes.string_to_list(K)
            sage: sdes.subkey(L, n=1)
            [1, 0, 1, 0, 0, 1, 0, 1]
            sage: sdes.subkey(sdes.string_to_list("0010010011"), n=2)
            [0, 1, 1, 0, 1, 0, 1, 0]

        TESTS:

        Input ``K`` must be a 10-bit key::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.subkey("K")
            Traceback (most recent call last):
            ...
            TypeError: input K must be a 10-bit key
            sage: sdes.subkey([])
            Traceback (most recent call last):
            ...
            ValueError: input K must be a 10-bit key

        There are only two subkeys::

            sage: key = [1, 0, 1, 0, 0, 0, 0, 0, 1, 0]
            sage: sdes.subkey(key, n=0)
            Traceback (most recent call last):
            ...
            ValueError: input n must be either 1 or 2
            sage: sdes.subkey(key, n=3)
            Traceback (most recent call last):
            ...
            ValueError: input n must be either 1 or 2
        """
        # sanity check
        if not isinstance(K, list):
            raise TypeError("input K must be a 10-bit key")
        if len(K) != self._key_size:
            raise ValueError("input K must be a 10-bit key")

        # get the first subkey
        if n == 1:
            key1 = self.permutation10(K)
            key1 = self.left_shift(key1, n=1)
            return self.permutation8(key1)
        # get the second subkey
        elif n == 2:
            key2 = self.permutation10(K)
            key2 = self.left_shift(key2, n=1)
            key2 = self.left_shift(key2, n=2)
            return self.permutation8(key2)
        # an invalid subkey number
        else:
            raise ValueError("input n must be either 1 or 2")

    def switch(self, B):
        r"""
        Interchange the first 4 bits with the last 4 bits in the list ``B``
        of 8 bits. Let `(b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7)`
        be a vector of 8 bits, where each `b_i \in \{ 0, 1 \}`. Then the
        switch function `\sigma` is given by

        .. MATH::

            \sigma(b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7)
            = (b_4, b_5, b_6, b_7, b_0, b_1, b_2, b_3)

        INPUT:

        - ``B`` -- list; a block of 8 bits

        OUTPUT:

        A block of the same dimension, but in which the first 4 bits from
        ``B`` has been switched for the last 4 bits in ``B``.

        EXAMPLES:

        Interchange the first 4 bits with the last 4 bits::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: B = [1, 1, 1, 0, 1, 0, 0, 0]
            sage: sdes.switch(B)
            [1, 0, 0, 0, 1, 1, 1, 0]
            sage: sdes.switch([1, 1, 1, 1, 0, 0, 0, 0])
            [0, 0, 0, 0, 1, 1, 1, 1]

        We can also work with a string of bits::

            sage: S = "11101000"
            sage: L = sdes.string_to_list(S)
            sage: sdes.switch(L)
            [1, 0, 0, 0, 1, 1, 1, 0]
            sage: sdes.switch(sdes.string_to_list("11110000"))
            [0, 0, 0, 0, 1, 1, 1, 1]

        TESTS:

        The input block must be a list::

            sage: from sage.crypto.block_cipher.sdes import SimplifiedDES
            sage: sdes = SimplifiedDES()
            sage: sdes.switch("B")
            Traceback (most recent call last):
            ...
            TypeError: input block must be a list of 8 bits
            sage: sdes.switch(())
            Traceback (most recent call last):
            ...
            TypeError: input block must be a list of 8 bits

        The input block must be a list of 8 bits::

            sage: sdes.switch([])
            Traceback (most recent call last):
            ...
            ValueError: input block must be a list of 8 bits
            sage: sdes.switch([1, 2, 3, 4, 5, 6, 7, 8, 9])
            Traceback (most recent call last):
            ...
            ValueError: input block must be a list of 8 bits

        The value of each element of the list must be either 0 or 1::

            sage: sdes.switch([1, 2, 3, 4, 5, 6, 7, 8])
            Traceback (most recent call last):
            ...
            TypeError: Argument x (= 5) is not a valid string.
        """
        # sanity check
        if not isinstance(B, list):
            raise TypeError("input block must be a list of 8 bits")
        if len(B) != 8:
            raise ValueError("input block must be a list of 8 bits")

        # perform the switch
        bin = BinaryStrings()
        return [ bin(str(B[4])), bin(str(B[5])),
                 bin(str(B[6])), bin(str(B[7])),
                 bin(str(B[0])), bin(str(B[1])),
                 bin(str(B[2])), bin(str(B[3])) ]



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/perm.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    slong * _perm_init(slong n) noexcept
    void _perm_clear(slong * vec) noexcept
    void _perm_set(slong * res, const slong * vec, slong n) noexcept
    void _perm_one(slong * vec, slong n) noexcept
    void _perm_inv(slong * res, const slong * vec, slong n) noexcept
    void _perm_compose(slong * res, const slong * vec1, const slong * vec2, slong n) noexcept
    int _perm_parity(const slong * vec, slong n) noexcept
    int _perm_randtest(slong * vec, slong n, flint_rand_t state) noexcept
    int _perm_print(const slong * vec, slong n) noexcept



# sage_setup: distribution = sagemath-modules
# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
from sage.libs.gsl.types cimport *

cdef extern from "gsl/gsl_fft.h":
  cdef enum gsl_fft_direction:
    forward = -1
    backward = 1

cdef extern from "gsl/gsl_fft_complex.h":

  # Power of 2 routines
  int  gsl_fft_complex_radix2_forward(gsl_complex_packed_array data, size_t stride, size_t n)

  int  gsl_fft_complex_radix2_backward(gsl_complex_packed_array data, size_t stride, size_t n)

  int  gsl_fft_complex_radix2_inverse(gsl_complex_packed_array data, size_t stride, size_t n)

  int  gsl_fft_complex_radix2_transform(gsl_complex_packed_array data, size_t stride, size_t n, gsl_fft_direction sign)

  int  gsl_fft_complex_radix2_dif_forward(gsl_complex_packed_array data, size_t stride, size_t n)

  int  gsl_fft_complex_radix2_dif_backward(gsl_complex_packed_array data, size_t stride, size_t n)

  int  gsl_fft_complex_radix2_dif_inverse(gsl_complex_packed_array data, size_t stride, size_t n)

  int  gsl_fft_complex_radix2_dif_transform(gsl_complex_packed_array data, size_t stride, size_t n, gsl_fft_direction sign)

  int gsl_fft_complex_bitreverse_order (gsl_complex_packed_array data, size_t stride, size_t n, size_t n_bits)


  # Mixed Radix general-N routines
  # or ctypedef struct gsl_fft_complex_wavetable
  ctypedef struct gsl_fft_complex_wavetable:
    size_t n
    size_t nf
    size_t factor[64]
    gsl_complex * trig
    gsl_complex * twiddle[64]

  ctypedef struct gsl_fft_complex_workspace

  gsl_fft_complex_wavetable *  gsl_fft_complex_wavetable_alloc(size_t n)

  void  gsl_fft_complex_wavetable_free(gsl_fft_complex_wavetable * wavetable)

  gsl_fft_complex_workspace *  gsl_fft_complex_workspace_alloc(size_t n)

  void  gsl_fft_complex_workspace_free(gsl_fft_complex_workspace * workspace)

  int gsl_fft_complex_memcpy (gsl_fft_complex_wavetable * dest, gsl_fft_complex_wavetable * src)

  int  gsl_fft_complex_forward(gsl_complex_packed_array data, size_t stride, size_t n, gsl_fft_complex_wavetable * wavetable, gsl_fft_complex_workspace * work)

  int  gsl_fft_complex_backward(gsl_complex_packed_array data, size_t stride, size_t n, gsl_fft_complex_wavetable * wavetable, gsl_fft_complex_workspace * work)

  int  gsl_fft_complex_inverse(gsl_complex_packed_array data, size_t stride, size_t n, gsl_fft_complex_wavetable * wavetable, gsl_fft_complex_workspace * work)

  int  gsl_fft_complex_transform(gsl_complex_packed_array data, size_t stride, size_t n, gsl_fft_complex_wavetable * wavetable, gsl_fft_complex_workspace * work, gsl_fft_direction sign)

  # end of gsl_fft_complex.h

cdef extern from "gsl/gsl_fft_real.h":

  int  gsl_fft_real_radix2_transform(double data[], size_t stride, size_t n)

  # or ctypedef struct gsl_fft_real_wavetable
  cdef struct gsl_fft_real_wavetable:
    size_t n
    size_t nf
    size_t factor[64]
    gsl_complex *twiddle[64]
    gsl_complex *trig

  cdef struct gsl_fft_real_workspace

  gsl_fft_real_wavetable *  gsl_fft_real_wavetable_alloc(size_t n)

  void  gsl_fft_real_wavetable_free(gsl_fft_real_wavetable * wavetable)

  gsl_fft_real_workspace *  gsl_fft_real_workspace_alloc(size_t n)

  void  gsl_fft_real_workspace_free(gsl_fft_real_workspace * workspace)

  int  gsl_fft_real_transform(double data[], size_t stride, size_t n, gsl_fft_real_wavetable * wavetable, gsl_fft_real_workspace * work)

  int  gsl_fft_real_unpack(double real_coefficient[], gsl_complex_packed_array complex_coefficient[], size_t stride, size_t n)

  # end of gsl_fft_real.h

cdef extern from "gsl/gsl_fft_halfcomplex.h":

  int  gsl_fft_halfcomplex_radix2_backward(double data[], size_t stride, size_t n)

  int  gsl_fft_halfcomplex_radix2_inverse(double data[], size_t stride, size_t n)

  int gsl_fft_halfcomplex_radix2_transform (double data[],  size_t stride,  size_t n)

  # or ctypedef struct gsl_fft_halfcomplex_wavetable
  cdef struct gsl_fft_halfcomplex_wavetable:
    size_t n
    size_t nf
    size_t factor[64]
    gsl_complex *twiddle[64]
    gsl_complex *trig

  gsl_fft_halfcomplex_wavetable *  gsl_fft_halfcomplex_wavetable_alloc(size_t n)

  void  gsl_fft_halfcomplex_wavetable_free(gsl_fft_halfcomplex_wavetable * wavetable)

  int gsl_fft_halfcomplex_backward (double data[],  size_t stride,  size_t n,  gsl_fft_halfcomplex_wavetable * wavetable, gsl_fft_real_workspace * work)

  int gsl_fft_halfcomplex_inverse (double data[],  size_t stride,  size_t n,  gsl_fft_halfcomplex_wavetable * wavetable, gsl_fft_real_workspace * work)

  int gsl_fft_halfcomplex_transform (double data[],  size_t stride,  size_t n,  gsl_fft_halfcomplex_wavetable * wavetable, gsl_fft_real_workspace * work)

  int  gsl_fft_halfcomplex_unpack(double halfcomplex_coefficient[], gsl_complex_packed_array complex_coefficient, size_t stride, size_t n)

  int gsl_fft_halfcomplex_radix2_unpack ( double halfcomplex_coefficient[], double complex_coefficient[],  size_t stride,  size_t n)



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/acb_dft.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void acb_dft(acb_ptr w, acb_srcptr v, slong n, slong prec) noexcept
    void acb_dft_inverse(acb_ptr w, acb_srcptr v, slong n, slong prec) noexcept
    void acb_dft_precomp_init(acb_dft_pre_t pre, slong len, slong prec) noexcept
    void acb_dft_precomp_clear(acb_dft_pre_t pre) noexcept
    void acb_dft_precomp(acb_ptr w, acb_srcptr v, const acb_dft_pre_t pre, slong prec) noexcept
    void acb_dft_inverse_precomp(acb_ptr w, acb_srcptr v, const acb_dft_pre_t pre, slong prec) noexcept
    void acb_dirichlet_dft_prod(acb_ptr w, acb_srcptr v, slong * cyc, slong num, slong prec) noexcept
    void acb_dft_prod_init(acb_dft_prod_t t, slong * cyc, slong num, slong prec) noexcept
    void acb_dft_prod_clear(acb_dft_prod_t t) noexcept
    void acb_dirichlet_dft_prod_precomp(acb_ptr w, acb_srcptr v, const acb_dft_prod_t prod, slong prec) noexcept
    void acb_dft_convol_naive(acb_ptr w, acb_srcptr f, acb_srcptr g, slong len, slong prec) noexcept
    void acb_dft_convol_rad2(acb_ptr w, acb_srcptr f, acb_srcptr g, slong len, slong prec) noexcept
    void acb_dft_convol(acb_ptr w, acb_srcptr f, acb_srcptr g, slong len, slong prec) noexcept
    void acb_dft_naive(acb_ptr w, acb_srcptr v, slong n, slong prec) noexcept
    void acb_dft_naive_init(acb_dft_naive_t t, slong len, slong prec) noexcept
    void acb_dft_naive_clear(acb_dft_naive_t t) noexcept
    void acb_dft_naive_precomp(acb_ptr w, acb_srcptr v, const acb_dft_naive_t t, slong prec) noexcept
    void acb_dft_crt(acb_ptr w, acb_srcptr v, slong n, slong prec) noexcept
    void acb_dft_crt_init(acb_dft_crt_t t, slong len, slong prec) noexcept
    void acb_dft_crt_clear(acb_dft_crt_t t) noexcept
    void acb_dft_crt_precomp(acb_ptr w, acb_srcptr v, const acb_dft_crt_t t, slong prec) noexcept
    void acb_dft_cyc(acb_ptr w, acb_srcptr v, slong n, slong prec) noexcept
    void acb_dft_cyc_init(acb_dft_cyc_t t, slong len, slong prec) noexcept
    void acb_dft_cyc_clear(acb_dft_cyc_t t) noexcept
    void acb_dft_cyc_precomp(acb_ptr w, acb_srcptr v, const acb_dft_cyc_t t, slong prec) noexcept
    void acb_dft_rad2(acb_ptr w, acb_srcptr v, int e, slong prec) noexcept
    void acb_dft_inverse_rad2(acb_ptr w, acb_srcptr v, int e, slong prec) noexcept
    void acb_dft_rad2_init(acb_dft_rad2_t t, int e, slong prec) noexcept
    void acb_dft_rad2_clear(acb_dft_rad2_t t) noexcept
    void acb_dft_rad2_precomp(acb_ptr w, acb_srcptr v, const acb_dft_rad2_t t, slong prec) noexcept
    void acb_dft_bluestein(acb_ptr w, acb_srcptr v, slong n, slong prec) noexcept
    void acb_dft_bluestein_init(acb_dft_bluestein_t t, slong len, slong prec) noexcept
    void acb_dft_bluestein_clear(acb_dft_bluestein_t t) noexcept
    void acb_dft_bluestein_precomp(acb_ptr w, acb_srcptr v, const acb_dft_bluestein_t t, slong prec) noexcept



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/acb_poly.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void acb_poly_init(acb_poly_t poly) noexcept
    void acb_poly_clear(acb_poly_t poly) noexcept
    void acb_poly_fit_length(acb_poly_t poly, slong len) noexcept
    void _acb_poly_set_length(acb_poly_t poly, slong len) noexcept
    void _acb_poly_normalise(acb_poly_t poly) noexcept
    void acb_poly_swap(acb_poly_t poly1, acb_poly_t poly2) noexcept
    slong acb_poly_allocated_bytes(const acb_poly_t x) noexcept
    slong acb_poly_length(const acb_poly_t poly) noexcept
    slong acb_poly_degree(const acb_poly_t poly) noexcept
    bint acb_poly_is_zero(const acb_poly_t poly) noexcept
    bint acb_poly_is_one(const acb_poly_t poly) noexcept
    bint acb_poly_is_x(const acb_poly_t poly) noexcept
    void acb_poly_zero(acb_poly_t poly) noexcept
    void acb_poly_one(acb_poly_t poly) noexcept
    void acb_poly_set(acb_poly_t dest, const acb_poly_t src) noexcept
    void acb_poly_set_round(acb_poly_t dest, const acb_poly_t src, slong prec) noexcept
    void acb_poly_set_trunc(acb_poly_t dest, const acb_poly_t src, slong n) noexcept
    void acb_poly_set_trunc_round(acb_poly_t dest, const acb_poly_t src, slong n, slong prec) noexcept
    void acb_poly_set_coeff_si(acb_poly_t poly, slong n, slong c) noexcept
    void acb_poly_set_coeff_acb(acb_poly_t poly, slong n, const acb_t c) noexcept
    void acb_poly_get_coeff_acb(acb_t v, const acb_poly_t poly, slong n) noexcept
    void _acb_poly_shift_right(acb_ptr res, acb_srcptr poly, slong len, slong n) noexcept
    void acb_poly_shift_right(acb_poly_t res, const acb_poly_t poly, slong n) noexcept
    void _acb_poly_shift_left(acb_ptr res, acb_srcptr poly, slong len, slong n) noexcept
    void acb_poly_shift_left(acb_poly_t res, const acb_poly_t poly, slong n) noexcept
    void acb_poly_truncate(acb_poly_t poly, slong n) noexcept
    slong acb_poly_valuation(const acb_poly_t poly) noexcept
    void acb_poly_printd(const acb_poly_t poly, slong digits) noexcept
    void acb_poly_fprintd(FILE * file, const acb_poly_t poly, slong digits) noexcept
    void acb_poly_randtest(acb_poly_t poly, flint_rand_t state, slong len, slong prec, slong mag_bits) noexcept
    bint acb_poly_equal(const acb_poly_t A, const acb_poly_t B) noexcept
    bint acb_poly_contains(const acb_poly_t poly1, const acb_poly_t poly2) noexcept
    bint acb_poly_contains_fmpz_poly(const acb_poly_t poly1, const fmpz_poly_t poly2) noexcept
    bint acb_poly_contains_fmpq_poly(const acb_poly_t poly1, const fmpq_poly_t poly2) noexcept
    bint _acb_poly_overlaps(acb_srcptr poly1, slong len1, acb_srcptr poly2, slong len2) noexcept
    bint acb_poly_overlaps(const acb_poly_t poly1, const acb_poly_t poly2) noexcept
    int acb_poly_get_unique_fmpz_poly(fmpz_poly_t z, const acb_poly_t x) noexcept
    bint acb_poly_is_real(const acb_poly_t poly) noexcept
    void acb_poly_set_fmpz_poly(acb_poly_t poly, const fmpz_poly_t re, slong prec) noexcept
    void acb_poly_set2_fmpz_poly(acb_poly_t poly, const fmpz_poly_t re, const fmpz_poly_t im, slong prec) noexcept
    void acb_poly_set_arb_poly(acb_poly_t poly, const arb_poly_t re) noexcept
    void acb_poly_set2_arb_poly(acb_poly_t poly, const arb_poly_t re, const arb_poly_t im) noexcept
    void acb_poly_set_fmpq_poly(acb_poly_t poly, const fmpq_poly_t re, slong prec) noexcept
    void acb_poly_set2_fmpq_poly(acb_poly_t poly, const fmpq_poly_t re, const fmpq_poly_t im, slong prec) noexcept
    void acb_poly_set_acb(acb_poly_t poly, const acb_t src) noexcept
    void acb_poly_set_si(acb_poly_t poly, slong src) noexcept
    void _acb_poly_majorant(arb_ptr res, acb_srcptr poly, slong len, slong prec) noexcept
    void acb_poly_majorant(arb_poly_t res, const acb_poly_t poly, slong prec) noexcept
    void _acb_poly_add(acb_ptr C, acb_srcptr A, slong lenA, acb_srcptr B, slong lenB, slong prec) noexcept
    void acb_poly_add(acb_poly_t C, const acb_poly_t A, const acb_poly_t B, slong prec) noexcept
    void acb_poly_add_si(acb_poly_t C, const acb_poly_t A, slong B, slong prec) noexcept
    void _acb_poly_sub(acb_ptr C, acb_srcptr A, slong lenA, acb_srcptr B, slong lenB, slong prec) noexcept
    void acb_poly_sub(acb_poly_t C, const acb_poly_t A, const acb_poly_t B, slong prec) noexcept
    void acb_poly_add_series(acb_poly_t C, const acb_poly_t A, const acb_poly_t B, slong len, slong prec) noexcept
    void acb_poly_sub_series(acb_poly_t C, const acb_poly_t A, const acb_poly_t B, slong len, slong prec) noexcept
    void acb_poly_neg(acb_poly_t C, const acb_poly_t A) noexcept
    void acb_poly_scalar_mul_2exp_si(acb_poly_t C, const acb_poly_t A, slong c) noexcept
    void acb_poly_scalar_mul(acb_poly_t C, const acb_poly_t A, const acb_t c, slong prec) noexcept
    void acb_poly_scalar_div(acb_poly_t C, const acb_poly_t A, const acb_t c, slong prec) noexcept
    void _acb_poly_mullow_classical(acb_ptr C, acb_srcptr A, slong lenA, acb_srcptr B, slong lenB, slong n, slong prec) noexcept
    void _acb_poly_mullow_transpose(acb_ptr C, acb_srcptr A, slong lenA, acb_srcptr B, slong lenB, slong n, slong prec) noexcept
    void _acb_poly_mullow_transpose_gauss(acb_ptr C, acb_srcptr A, slong lenA, acb_srcptr B, slong lenB, slong n, slong prec) noexcept
    void _acb_poly_mullow(acb_ptr C, acb_srcptr A, slong lenA, acb_srcptr B, slong lenB, slong n, slong prec) noexcept
    void acb_poly_mullow_classical(acb_poly_t C, const acb_poly_t A, const acb_poly_t B, slong n, slong prec) noexcept
    void acb_poly_mullow_transpose(acb_poly_t C, const acb_poly_t A, const acb_poly_t B, slong n, slong prec) noexcept
    void acb_poly_mullow_transpose_gauss(acb_poly_t C, const acb_poly_t A, const acb_poly_t B, slong n, slong prec) noexcept
    void acb_poly_mullow(acb_poly_t C, const acb_poly_t A, const acb_poly_t B, slong n, slong prec) noexcept
    void _acb_poly_mul(acb_ptr C, acb_srcptr A, slong lenA, acb_srcptr B, slong lenB, slong prec) noexcept
    void acb_poly_mul(acb_poly_t C, const acb_poly_t A1, const acb_poly_t B2, slong prec) noexcept
    void _acb_poly_inv_series(acb_ptr Qinv, acb_srcptr Q, slong Qlen, slong len, slong prec) noexcept
    void acb_poly_inv_series(acb_poly_t Qinv, const acb_poly_t Q, slong n, slong prec) noexcept
    void _acb_poly_div_series(acb_ptr Q, acb_srcptr A, slong Alen, acb_srcptr B, slong Blen, slong n, slong prec) noexcept
    void acb_poly_div_series(acb_poly_t Q, const acb_poly_t A, const acb_poly_t B, slong n, slong prec) noexcept
    void _acb_poly_div(acb_ptr Q, acb_srcptr A, slong lenA, acb_srcptr B, slong lenB, slong prec) noexcept
    void _acb_poly_rem(acb_ptr R, acb_srcptr A, slong lenA, acb_srcptr B, slong lenB, slong prec) noexcept
    void _acb_poly_divrem(acb_ptr Q, acb_ptr R, acb_srcptr A, slong lenA, acb_srcptr B, slong lenB, slong prec) noexcept
    int acb_poly_divrem(acb_poly_t Q, acb_poly_t R, const acb_poly_t A, const acb_poly_t B, slong prec) noexcept
    void _acb_poly_div_root(acb_ptr Q, acb_t R, acb_srcptr A, slong len, const acb_t c, slong prec) noexcept
    void _acb_poly_taylor_shift(acb_ptr g, const acb_t c, slong n, slong prec) noexcept
    void acb_poly_taylor_shift(acb_poly_t g, const acb_poly_t f, const acb_t c, slong prec) noexcept
    void _acb_poly_compose(acb_ptr res, acb_srcptr poly1, slong len1, acb_srcptr poly2, slong len2, slong prec) noexcept
    void acb_poly_compose(acb_poly_t res, const acb_poly_t poly1, const acb_poly_t poly2, slong prec) noexcept
    void _acb_poly_compose_series(acb_ptr res, acb_srcptr poly1, slong len1, acb_srcptr poly2, slong len2, slong n, slong prec) noexcept
    void acb_poly_compose_series(acb_poly_t res, const acb_poly_t poly1, const acb_poly_t poly2, slong n, slong prec) noexcept
    void _acb_poly_revert_series(acb_ptr h, acb_srcptr f, slong flen, slong n, slong prec) noexcept
    void acb_poly_revert_series(acb_poly_t h, const acb_poly_t f, slong n, slong prec) noexcept
    void _acb_poly_evaluate_horner(acb_t y, acb_srcptr f, slong len, const acb_t x, slong prec) noexcept
    void acb_poly_evaluate_horner(acb_t y, const acb_poly_t f, const acb_t x, slong prec) noexcept
    void _acb_poly_evaluate_rectangular(acb_t y, acb_srcptr f, slong len, const acb_t x, slong prec) noexcept
    void acb_poly_evaluate_rectangular(acb_t y, const acb_poly_t f, const acb_t x, slong prec) noexcept
    void _acb_poly_evaluate(acb_t y, acb_srcptr f, slong len, const acb_t x, slong prec) noexcept
    void acb_poly_evaluate(acb_t y, const acb_poly_t f, const acb_t x, slong prec) noexcept
    void _acb_poly_evaluate2_horner(acb_t y, acb_t z, acb_srcptr f, slong len, const acb_t x, slong prec) noexcept
    void acb_poly_evaluate2_horner(acb_t y, acb_t z, const acb_poly_t f, const acb_t x, slong prec) noexcept
    void _acb_poly_evaluate2_rectangular(acb_t y, acb_t z, acb_srcptr f, slong len, const acb_t x, slong prec) noexcept
    void acb_poly_evaluate2_rectangular(acb_t y, acb_t z, const acb_poly_t f, const acb_t x, slong prec) noexcept
    void _acb_poly_evaluate2(acb_t y, acb_t z, acb_srcptr f, slong len, const acb_t x, slong prec) noexcept
    void acb_poly_evaluate2(acb_t y, acb_t z, const acb_poly_t f, const acb_t x, slong prec) noexcept
    void _acb_poly_product_roots(acb_ptr poly, acb_srcptr xs, slong n, slong prec) noexcept
    void acb_poly_product_roots(acb_poly_t poly, acb_srcptr xs, slong n, slong prec) noexcept
    acb_ptr * _acb_poly_tree_alloc(slong len) noexcept
    void _acb_poly_tree_free(acb_ptr * tree, slong len) noexcept
    void _acb_poly_tree_build(acb_ptr * tree, acb_srcptr roots, slong len, slong prec) noexcept
    void _acb_poly_evaluate_vec_iter(acb_ptr ys, acb_srcptr poly, slong plen, acb_srcptr xs, slong n, slong prec) noexcept
    void acb_poly_evaluate_vec_iter(acb_ptr ys, const acb_poly_t poly, acb_srcptr xs, slong n, slong prec) noexcept
    void _acb_poly_evaluate_vec_fast_precomp(acb_ptr vs, acb_srcptr poly, slong plen, acb_ptr * tree, slong len, slong prec) noexcept
    void _acb_poly_evaluate_vec_fast(acb_ptr ys, acb_srcptr poly, slong plen, acb_srcptr xs, slong n, slong prec) noexcept
    void acb_poly_evaluate_vec_fast(acb_ptr ys, const acb_poly_t poly, acb_srcptr xs, slong n, slong prec) noexcept
    void _acb_poly_interpolate_newton(acb_ptr poly, acb_srcptr xs, acb_srcptr ys, slong n, slong prec) noexcept
    void acb_poly_interpolate_newton(acb_poly_t poly, acb_srcptr xs, acb_srcptr ys, slong n, slong prec) noexcept
    void _acb_poly_interpolate_barycentric(acb_ptr poly, acb_srcptr xs, acb_srcptr ys, slong n, slong prec) noexcept
    void acb_poly_interpolate_barycentric(acb_poly_t poly, acb_srcptr xs, acb_srcptr ys, slong n, slong prec) noexcept
    void _acb_poly_interpolation_weights(acb_ptr w, acb_ptr * tree, slong len, slong prec) noexcept
    void _acb_poly_interpolate_fast_precomp(acb_ptr poly, acb_srcptr ys, acb_ptr * tree, acb_srcptr weights, slong len, slong prec) noexcept
    void _acb_poly_interpolate_fast(acb_ptr poly, acb_srcptr xs, acb_srcptr ys, slong len, slong prec) noexcept
    void acb_poly_interpolate_fast(acb_poly_t poly, acb_srcptr xs, acb_srcptr ys, slong n, slong prec) noexcept
    void _acb_poly_derivative(acb_ptr res, acb_srcptr poly, slong len, slong prec) noexcept
    void acb_poly_derivative(acb_poly_t res, const acb_poly_t poly, slong prec) noexcept
    void _acb_poly_nth_derivative(acb_ptr res, acb_srcptr poly, ulong n, slong len, slong prec) noexcept
    void acb_poly_nth_derivative(acb_poly_t res, const acb_poly_t poly, ulong n, slong prec) noexcept
    void _acb_poly_integral(acb_ptr res, acb_srcptr poly, slong len, slong prec) noexcept
    void acb_poly_integral(acb_poly_t res, const acb_poly_t poly, slong prec) noexcept
    void _acb_poly_borel_transform(acb_ptr res, acb_srcptr poly, slong len, slong prec) noexcept
    void acb_poly_borel_transform(acb_poly_t res, const acb_poly_t poly, slong prec) noexcept
    void _acb_poly_inv_borel_transform(acb_ptr res, acb_srcptr poly, slong len, slong prec) noexcept
    void acb_poly_inv_borel_transform(acb_poly_t res, const acb_poly_t poly, slong prec) noexcept
    void _acb_poly_binomial_transform_basecase(acb_ptr b, acb_srcptr a, slong alen, slong len, slong prec) noexcept
    void acb_poly_binomial_transform_basecase(acb_poly_t b, const acb_poly_t a, slong len, slong prec) noexcept
    void _acb_poly_binomial_transform_convolution(acb_ptr b, acb_srcptr a, slong alen, slong len, slong prec) noexcept
    void acb_poly_binomial_transform_convolution(acb_poly_t b, const acb_poly_t a, slong len, slong prec) noexcept
    void _acb_poly_binomial_transform(acb_ptr b, acb_srcptr a, slong alen, slong len, slong prec) noexcept
    void acb_poly_binomial_transform(acb_poly_t b, const acb_poly_t a, slong len, slong prec) noexcept
    void _acb_poly_graeffe_transform(acb_ptr b, acb_srcptr a, slong len, slong prec) noexcept
    void acb_poly_graeffe_transform(acb_poly_t b, const acb_poly_t a, slong prec) noexcept
    void _acb_poly_pow_ui_trunc_binexp(acb_ptr res, acb_srcptr f, slong flen, ulong exp, slong len, slong prec) noexcept
    void acb_poly_pow_ui_trunc_binexp(acb_poly_t res, const acb_poly_t poly, ulong exp, slong len, slong prec) noexcept
    void _acb_poly_pow_ui(acb_ptr res, acb_srcptr f, slong flen, ulong exp, slong prec) noexcept
    void acb_poly_pow_ui(acb_poly_t res, const acb_poly_t poly, ulong exp, slong prec) noexcept
    void _acb_poly_pow_series(acb_ptr h, acb_srcptr f, slong flen, acb_srcptr g, slong glen, slong len, slong prec) noexcept
    void acb_poly_pow_series(acb_poly_t h, const acb_poly_t f, const acb_poly_t g, slong len, slong prec) noexcept
    void _acb_poly_pow_acb_series(acb_ptr h, acb_srcptr f, slong flen, const acb_t g, slong len, slong prec) noexcept
    void acb_poly_pow_acb_series(acb_poly_t h, const acb_poly_t f, const acb_t g, slong len, slong prec) noexcept
    void _acb_poly_sqrt_series(acb_ptr g, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_sqrt_series(acb_poly_t g, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_rsqrt_series(acb_ptr g, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_rsqrt_series(acb_poly_t g, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_log_series(acb_ptr res, acb_srcptr f, slong flen, slong n, slong prec) noexcept
    void acb_poly_log_series(acb_poly_t res, const acb_poly_t f, slong n, slong prec) noexcept
    void _acb_poly_log1p_series(acb_ptr res, acb_srcptr f, slong flen, slong n, slong prec) noexcept
    void acb_poly_log1p_series(acb_poly_t res, const acb_poly_t f, slong n, slong prec) noexcept
    void _acb_poly_atan_series(acb_ptr res, acb_srcptr f, slong flen, slong n, slong prec) noexcept
    void acb_poly_atan_series(acb_poly_t res, const acb_poly_t f, slong n, slong prec) noexcept
    void _acb_poly_exp_series_basecase(acb_ptr f, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_exp_series_basecase(acb_poly_t f, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_exp_series(acb_ptr f, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_exp_series(acb_poly_t f, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_exp_pi_i_series(acb_ptr f, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_exp_pi_i_series(acb_poly_t f, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_sin_cos_series(acb_ptr s, acb_ptr c, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_sin_cos_series(acb_poly_t s, acb_poly_t c, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_sin_series(acb_ptr s, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_sin_series(acb_poly_t s, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_cos_series(acb_ptr c, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_cos_series(acb_poly_t c, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_tan_series(acb_ptr g, acb_srcptr h, slong hlen, slong len, slong prec) noexcept
    void acb_poly_tan_series(acb_poly_t g, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_sin_cos_pi_series(acb_ptr s, acb_ptr c, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_sin_cos_pi_series(acb_poly_t s, acb_poly_t c, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_sin_pi_series(acb_ptr s, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_sin_pi_series(acb_poly_t s, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_cos_pi_series(acb_ptr c, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_cos_pi_series(acb_poly_t c, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_cot_pi_series(acb_ptr c, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_cot_pi_series(acb_poly_t c, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_sinh_cosh_series_basecase(acb_ptr s, acb_ptr c, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_sinh_cosh_series_basecase(acb_poly_t s, acb_poly_t c, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_sinh_cosh_series_exponential(acb_ptr s, acb_ptr c, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_sinh_cosh_series_exponential(acb_poly_t s, acb_poly_t c, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_sinh_cosh_series(acb_ptr s, acb_ptr c, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_sinh_cosh_series(acb_poly_t s, acb_poly_t c, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_sinh_series(acb_ptr s, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_sinh_series(acb_poly_t s, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_cosh_series(acb_ptr c, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_cosh_series(acb_poly_t c, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_sinc_series(acb_ptr s, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_sinc_series(acb_poly_t s, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_lambertw_series(acb_ptr res, acb_srcptr z, slong zlen, const fmpz_t k, int flags, slong len, slong prec) noexcept
    void acb_poly_lambertw_series(acb_poly_t res, const acb_poly_t z, const fmpz_t k, int flags, slong len, slong prec) noexcept
    void _acb_poly_gamma_series(acb_ptr res, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_gamma_series(acb_poly_t res, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_rgamma_series(acb_ptr res, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_rgamma_series(acb_poly_t res, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_lgamma_series(acb_ptr res, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_lgamma_series(acb_poly_t res, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_digamma_series(acb_ptr res, acb_srcptr h, slong hlen, slong n, slong prec) noexcept
    void acb_poly_digamma_series(acb_poly_t res, const acb_poly_t h, slong n, slong prec) noexcept
    void _acb_poly_rising_ui_series(acb_ptr res, acb_srcptr f, slong flen, ulong r, slong trunc, slong prec) noexcept
    void acb_poly_rising_ui_series(acb_poly_t res, const acb_poly_t f, ulong r, slong trunc, slong prec) noexcept
    void _acb_poly_powsum_series_naive(acb_ptr z, const acb_t s, const acb_t a, const acb_t q, slong n, slong len, slong prec) noexcept
    void _acb_poly_powsum_series_naive_threaded(acb_ptr z, const acb_t s, const acb_t a, const acb_t q, slong n, slong len, slong prec) noexcept
    void _acb_poly_powsum_one_series_sieved(acb_ptr z, const acb_t s, slong n, slong len, slong prec) noexcept
    void _acb_poly_zeta_em_choose_param(mag_t bound, ulong * N, ulong * M, const acb_t s, const acb_t a, slong d, slong target, slong prec) noexcept
    void _acb_poly_zeta_em_bound1(mag_t bound, const acb_t s, const acb_t a, slong N, slong M, slong d, slong wp) noexcept
    void _acb_poly_zeta_em_bound(arb_ptr vec, const acb_t s, const acb_t a, ulong N, ulong M, slong d, slong wp) noexcept
    void _acb_poly_zeta_em_tail_naive(acb_ptr z, const acb_t s, const acb_t Na, acb_srcptr Nasx, slong M, slong len, slong prec) noexcept
    void _acb_poly_zeta_em_tail_bsplit(acb_ptr z, const acb_t s, const acb_t Na, acb_srcptr Nasx, slong M, slong len, slong prec) noexcept
    void _acb_poly_zeta_em_sum(acb_ptr z, const acb_t s, const acb_t a, int deflate, ulong N, ulong M, slong d, slong prec) noexcept
    void _acb_poly_zeta_cpx_series(acb_ptr z, const acb_t s, const acb_t a, int deflate, slong d, slong prec) noexcept
    void _acb_poly_zeta_series(acb_ptr res, acb_srcptr h, slong hlen, const acb_t a, int deflate, slong len, slong prec) noexcept
    void acb_poly_zeta_series(acb_poly_t res, const acb_poly_t f, const acb_t a, int deflate, slong n, slong prec) noexcept
    void _acb_poly_polylog_cpx_small(acb_ptr w, const acb_t s, const acb_t z, slong len, slong prec) noexcept
    void _acb_poly_polylog_cpx_zeta(acb_ptr w, const acb_t s, const acb_t z, slong len, slong prec) noexcept
    void _acb_poly_polylog_cpx(acb_ptr w, const acb_t s, const acb_t z, slong len, slong prec) noexcept
    void _acb_poly_polylog_series(acb_ptr w, acb_srcptr s, slong slen, const acb_t z, slong len, slong prec) noexcept
    void acb_poly_polylog_series(acb_poly_t w, const acb_poly_t s, const acb_t z, slong len, slong prec) noexcept
    void _acb_poly_erf_series(acb_ptr res, acb_srcptr z, slong zlen, slong n, slong prec) noexcept
    void acb_poly_erf_series(acb_poly_t res, const acb_poly_t z, slong n, slong prec) noexcept
    void _acb_poly_agm1_series(acb_ptr res, acb_srcptr z, slong zlen, slong len, slong prec) noexcept
    void acb_poly_agm1_series(acb_poly_t res, const acb_poly_t z, slong n, slong prec) noexcept
    void _acb_poly_elliptic_k_series(acb_ptr res, acb_srcptr z, slong zlen, slong len, slong prec) noexcept
    void acb_poly_elliptic_k_series(acb_poly_t res, const acb_poly_t z, slong n, slong prec) noexcept
    void _acb_poly_elliptic_p_series(acb_ptr res, acb_srcptr z, slong zlen, const acb_t tau, slong len, slong prec) noexcept
    void acb_poly_elliptic_p_series(acb_poly_t res, const acb_poly_t z, const acb_t tau, slong n, slong prec) noexcept
    void _acb_poly_root_bound_fujiwara(mag_t bound, acb_srcptr poly, slong len) noexcept
    void acb_poly_root_bound_fujiwara(mag_t bound, acb_poly_t poly) noexcept
    void _acb_poly_root_inclusion(acb_t r, const acb_t m, acb_srcptr poly, acb_srcptr polyder, slong len, slong prec) noexcept
    slong _acb_poly_validate_roots(acb_ptr roots, acb_srcptr poly, slong len, slong prec) noexcept
    void _acb_poly_refine_roots_durand_kerner(acb_ptr roots, acb_srcptr poly, slong len, slong prec) noexcept
    slong _acb_poly_find_roots(acb_ptr roots, acb_srcptr poly, acb_srcptr initial, slong len, slong maxiter, slong prec) noexcept
    slong acb_poly_find_roots(acb_ptr roots, const acb_poly_t poly, acb_srcptr initial, slong maxiter, slong prec) noexcept
    int _acb_poly_validate_real_roots(acb_srcptr roots, acb_srcptr poly, slong len, slong prec) noexcept
    int acb_poly_validate_real_roots(acb_srcptr roots, const acb_poly_t poly, slong prec) noexcept

from .acb_poly_macros cimport *



# sage_setup: distribution = sagemath-pari
from cypari2.gen cimport Gen

from sage.rings.real_mpfr cimport RealNumber

cpdef Gen new_gen_from_real_mpfr_element(RealNumber self)
cpdef bint set_real_mpfr_element_from_gen(RealNumber self, Gen x) noexcept



# sage_setup: distribution = sagemath-modules
"""
Local Density Congruence
"""
##########################################################################
#  Methods which compute the local densities for representing a number
#  by a quadratic form at a prime (possibly subject to additional
#  congruence conditions).
##########################################################################
from copy import deepcopy

from sage.sets.set import Set
from sage.rings.rational_field import QQ
from sage.arith.misc import valuation
from sage.misc.verbose import verbose

from sage.quadratic_forms.count_local_2 import count_modp__by_gauss_sum, count_all_local_good_types_normal_form


def count_modp_solutions__by_Gauss_sum(self, p, m):
    r"""
    Return the number of solutions of `Q(x) = m` (mod `p`) of a
    non-degenerate quadratic form over the finite field `\ZZ/p\ZZ`,
    where `p` is a prime number > 2.

    .. NOTE::

        We adopt the useful convention that a zero-dimensional
        quadratic form has exactly one solution always (i.e. the empty
        vector).

    These are defined in Table 1 on p363 of Hanke's "Local Densities..." paper.

    INPUT:

    - ``p`` -- a prime number > 2

    - ``m`` -- integer

    OUTPUT: integer `\geq 0`

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: [Q.count_modp_solutions__by_Gauss_sum(3, m)  for m in range(3)]
        [9, 6, 12]

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,2])
        sage: [Q.count_modp_solutions__by_Gauss_sum(3, m)  for m in range(3)]
        [9, 12, 6]
    """
    if self.dim() == 0:
        return 1
    return count_modp__by_gauss_sum(self.dim(), p, m, self.Gram_det())


def local_good_density_congruence_odd(self, p, m, Zvec, NZvec):
    """
    Find the Good-type local density of `Q` representing `m` at `p`.
    (Assuming that `p > 2` and `Q` is given in local diagonal form.)

    The additional congruence condition arguments ``Zvec`` and ``NZvec`` can
    be either a list of indices or None.  ``Zvec=[]`` is equivalent to
    ``Zvec=None``, which both impose no additional conditions, but
    ``NZvec=[]`` returns no solutions always while ``NZvec=None`` imposes no
    additional condition.

    .. TODO::

        Add type checking for ``Zvec``, ``NZvec``, and that `Q` is in local
        normal form.

    INPUT:

    - ``self`` -- quadratic form `Q`, assumed to be diagonal and `p`-integral

    - ``p`` -- a prime number

    - ``m`` -- integer

    - ``Zvec``, ``NZvec`` -- non-repeating lists of integers in ``range(self.dim())`` or ``None``

    OUTPUT: a rational number

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.local_good_density_congruence_odd(3, 1, None, None)
        2/3

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q.local_good_density_congruence_odd(3, 1, None, None)
        8/9
    """
    n = self.dim()

    #  Put the Zvec congruence condition in a standard form
    if Zvec is None:
        Zvec = []

    #  Sanity Check on Zvec and NZvec:
    #  -------------------------------
    Sn = Set(range(n))
    if (Zvec is not None) and (len(Set(Zvec) + Sn) > n):
        raise RuntimeError("Zvec must be a subset of {0, ..., n-1}.")
    if (NZvec is not None) and (len(Set(NZvec) + Sn) > n):
        raise RuntimeError("NZvec must be a subset of {0, ..., n-1}.")

    #  Assuming Q is diagonal, find the indices of the p-unit (diagonal) entries
    UnitVec = Set(i for i in range(n) if self[i, i] % p)
    NonUnitVec = Set(range(n)) - UnitVec

    #  Take cases on the existence of additional nonzero congruence conditions (mod p)
    UnitVec_minus_Zvec = list(UnitVec - Set(Zvec))
    NonUnitVec_minus_Zvec = list(NonUnitVec - Set(Zvec))
    Q_Unit_minus_Zvec = self.extract_variables(UnitVec_minus_Zvec)

    if NZvec is None:
        if m % p:
            total = Q_Unit_minus_Zvec.count_modp_solutions__by_Gauss_sum(p, m) * p**len(NonUnitVec_minus_Zvec)
        else:
            total = (Q_Unit_minus_Zvec.count_modp_solutions__by_Gauss_sum(p, m) - 1) * p**len(NonUnitVec_minus_Zvec)

    else:
        UnitVec_minus_ZNZvec = list(UnitVec - (Set(Zvec) + Set(NZvec)))
        NonUnitVec_minus_ZNZvec = list(NonUnitVec - (Set(Zvec) + Set(NZvec)))
        Q_Unit_minus_ZNZvec = self.extract_variables(UnitVec_minus_ZNZvec)

        if m % p:
            total = Q_Unit_minus_Zvec.count_modp_solutions__by_Gauss_sum(p, m) * p**len(NonUnitVec_minus_Zvec) \
                - Q_Unit_minus_ZNZvec.count_modp_solutions__by_Gauss_sum(p, m) * p**len(NonUnitVec_minus_ZNZvec)
        else:
            total = (Q_Unit_minus_Zvec.count_modp_solutions__by_Gauss_sum(p, m) - 1) * p**len(NonUnitVec_minus_Zvec) \
                - (Q_Unit_minus_ZNZvec.count_modp_solutions__by_Gauss_sum(p, m) - 1) * p**len(NonUnitVec_minus_ZNZvec)

    #  Return the Good-type representation density
    good_density = QQ(total) / p**(n - 1)
    return good_density


def local_good_density_congruence_even(self, m, Zvec, NZvec):
    r"""
    Find the Good-type local density of `Q` representing `m` at `p=2`.
    (Assuming `Q` is given in local diagonal form.)

    The additional congruence condition arguments ``Zvec`` and ``NZvec`` can
    be either a list of indices or None.  ``Zvec=[]`` is equivalent to
    ``Zvec=None`` which both impose no additional conditions, but
    ``NZvec=[]`` returns no solutions always while ``NZvec=None`` imposes no
    additional condition.

    .. WARNING::

        Here the indices passed in ``Zvec`` and ``NZvec`` represent
        indices of the solution vector `x` of `Q(x) = m` (mod `p^k`), and *not*
        the Jordan components of `Q`.  They therefore are required (and
        assumed) to include either all or none of the indices of a given
        Jordan component of `Q`.  This is only important when `p=2` since
        otherwise all Jordan blocks are `1 \times 1`, and so there the indices and
        Jordan blocks coincide.

    .. TODO::

        Add type checking for ``Zvec`` and ``NZvec``, and that `Q` is in local
        normal form.

    INPUT:

    - ``self`` -- quadratic form `Q`, assumed to be block diagonal and 2-integral

    - ``p`` -- a prime number

    - ``m`` -- integer

    - ``Zvec``, ``NZvec`` -- non-repeating lists of integers in ``range(self.dim())`` or ``None``

    OUTPUT: a rational number

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.local_good_density_congruence_even(1, None, None)
        1

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q.local_good_density_congruence_even(1, None, None)
        1
        sage: Q.local_good_density_congruence_even(2, None, None)
        3/2
        sage: Q.local_good_density_congruence_even(3, None, None)
        1
        sage: Q.local_good_density_congruence_even(4, None, None)
        1/2

    ::

        sage: Q = QuadraticForm(ZZ, 4, range(10))
        sage: Q[0,0] = 5
        sage: Q[1,1] = 10
        sage: Q[2,2] = 15
        sage: Q[3,3] = 20
        sage: Q
        Quadratic form in 4 variables over Integer Ring with coefficients:
        [ 5 1 2 3 ]
        [ * 10 5 6 ]
        [ * * 15 8 ]
        [ * * * 20 ]
        sage: Q.theta_series(20)                                                        # needs sage.libs.pari
        1 + 2*q^5 + 2*q^10 + 2*q^14 + 2*q^15 + 2*q^16 + 2*q^18 + O(q^20)
        sage: Q_local = Q.local_normal_form(2)                                          # needs sage.libs.pari sage.rings.padics
        sage: Q_local.local_good_density_congruence_even(1, None, None)                 # needs sage.libs.pari sage.rings.padics
        3/4
        sage: Q_local.local_good_density_congruence_even(2, None, None)                 # needs sage.libs.pari sage.rings.padics
        9/8
        sage: Q_local.local_good_density_congruence_even(5, None, None)                 # needs sage.libs.pari sage.rings.padics
        3/4
    """
    n = self.dim()

    #  Put the Zvec congruence condition in a standard form
    if Zvec is None:
        Zvec = []

    #  Sanity Check on Zvec and NZvec:
    #  -------------------------------
    Sn = Set(range(n))
    if (Zvec is not None) and (len(Set(Zvec) + Sn) > n):
        raise RuntimeError("Zvec must be a subset of {0, ..., n-1}.")
    if (NZvec is not None) and (len(Set(NZvec) + Sn) > n):
        raise RuntimeError("NZvec must be a subset of {0, ..., n-1}.")

    #  Find the indices of x for which the associated Jordan blocks are nonzero mod 8    TODO: Move this to special Jordan block code separately!
    #  -------------------------------------------------------------------------------
    Not8vec = []
    for i in range(n):

        #  DIAGNOSTIC
        verbose(" i = " + str(i))
        verbose(" n = " + str(n))
        verbose(" Not8vec = " + str(Not8vec))

        nz_flag = False

        #  Check if the diagonal entry isn't divisible 8
        if self[i, i] % 8:
            nz_flag = True

        #  Check appropriate off-diagonal entries aren't divisible by 8
        else:

            #  Special check for first off-diagonal entry
            if i == 0 and self[i, i + 1] % 8:
                nz_flag = True

            #  Special check for last off-diagonal entry
            elif i == n - 1 and self[i - 1, i] % 8:
                nz_flag = True

            #  Check for the middle off-diagonal entries
            else:
                if (i > 0) and (i < n - 1) and (self[i, i + 1] % 8 or
                                                self[i - 1, i] % 8):
                    nz_flag = True

        #  Remember the (vector) index if it's not part of a Jordan block of norm divisible by 8
        if nz_flag:
            Not8vec += [i]

    #  Compute the number of Good-type solutions mod 8:
    #  ------------------------------------------------

    #  Setup the indexing sets for additional zero congruence solutions
    Q_Not8 = self.extract_variables(Not8vec)
    Not8 = Set(Not8vec)
    Is8 = Set(range(n)) - Not8

    Z = Set(Zvec)
    Z_Not8 = Not8.intersection(Z)
    Z_Is8 = Is8.intersection(Z)
    Is8_minus_Z = Is8 - Z_Is8

    # DIAGNOSTIC
    verbose("Z = " + str(Z))
    verbose("Z_Not8 = " + str(Z_Not8))
    verbose("Z_Is8 = " + str(Z_Is8))
    verbose("Is8_minus_Z = " + str(Is8_minus_Z))

    # Take cases on the existence of additional nonzero congruence conditions (mod 2)
    if NZvec is None:
        total = (4 ** len(Z_Is8)) * (8 ** len(Is8_minus_Z)) \
            * count_all_local_good_types_normal_form(Q_Not8, 2, 3, m, list(Z_Not8), None)
    else:
        ZNZ = Z + Set(NZvec)
        ZNZ_Not8 = Not8.intersection(ZNZ)
        ZNZ_Is8 = Is8.intersection(ZNZ)
        Is8_minus_ZNZ = Is8 - ZNZ_Is8

        #  DIAGNOSTIC
        verbose("ZNZ = " + str(ZNZ))
        verbose("ZNZ_Not8 = " + str(ZNZ_Not8))
        verbose("ZNZ_Is8 = " + str(ZNZ_Is8))
        verbose("Is8_minus_ZNZ = " + str(Is8_minus_ZNZ))

        total = (4 ** len(Z_Is8)) * (8 ** len(Is8_minus_Z)) \
            * count_all_local_good_types_normal_form(Q_Not8, 2, 3, m, list(Z_Not8), None) \
            - (4 ** len(ZNZ_Is8)) * (8 ** len(Is8_minus_ZNZ)) \
            * count_all_local_good_types_normal_form(Q_Not8, 2, 3, m, list(ZNZ_Not8), None)

    # DIAGNOSTIC
    verbose("total = " + str(total))

    # Return the associated Good-type representation density
    return QQ(total) / 8**(n - 1)


def local_good_density_congruence(self, p, m, Zvec=None, NZvec=None):
    """
    Find the Good-type local density of `Q` representing `m` at `p`.
    (Front end routine for parity specific routines for `p`.)

    .. TODO::

        Add documentation about the additional congruence
        conditions ``Zvec`` and ``NZvec``.

    INPUT:

    - ``self`` -- quadratic form `Q`, assumed to be block diagonal and `p`-integral

    - ``p`` -- a prime number

    - ``m`` -- integer

    - ``Zvec``, ``NZvec`` -- non-repeating lists of integers in ``range(self.dim())`` or ``None``

    OUTPUT: a rational number

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.local_good_density_congruence(2, 1, None, None)
        1
        sage: Q.local_good_density_congruence(3, 1, None, None)
        2/3

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q.local_good_density_congruence(2, 1, None, None)
        1
        sage: Q.local_good_density_congruence(3, 1, None, None)
        8/9
    """
    #  DIAGNOSTIC
    verbose(" In local_good_density_congruence with ")
    verbose(" Q is: \n" + str(self))
    verbose(" p = " + str(p))
    verbose(" m = " + str(m))
    verbose(" Zvec = " + str(Zvec))
    verbose(" NZvec = " + str(NZvec))

    #  Put the Zvec congruence condition in a standard form
    if Zvec is None:
        Zvec = []

    n = self.dim()

    #  Sanity Check on Zvec and NZvec:
    #  -------------------------------
    Sn = Set(range(n))
    if (Zvec is not None) and (len(Set(Zvec) + Sn) > n):
        raise RuntimeError("Zvec must be a subset of {0, ..., n-1}.")
    if (NZvec is not None) and (len(Set(NZvec) + Sn) > n):
        raise RuntimeError("NZvec must be a subset of {0, ..., n-1}.")

    # There was here a commented-out check that Q is in local normal form
    # (it often may not be since the reduction procedure
    # often mixes up the order of the valuations...)
    # This commented-out code was removed in issue #32960

    # Decide which routine to use to compute the Good-type density
    if p > 2:
        return self.local_good_density_congruence_odd(p, m, Zvec, NZvec)

    if p == 2:
        return self.local_good_density_congruence_even(m, Zvec, NZvec)

    raise RuntimeError("\n Error in Local_Good_Density: The 'prime' p = " + str(p) + " is < 2. \n")


def local_zero_density_congruence(self, p, m, Zvec=None, NZvec=None):
    """
    Find the Zero-type local density of `Q` representing `m` at `p`,
    allowing certain congruence conditions mod `p`.

    INPUT:

    - ``self`` -- quadratic form `Q`, assumed to be block diagonal and `p`-integral

    - ``p`` -- a prime number

    - ``m`` -- integer

    - ``Zvec``, ``NZvec`` -- non-repeating lists of integers in ``range(self.dim())`` or ``None``

    OUTPUT: a rational number

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.local_zero_density_congruence(2, 2, None, None)
        0
        sage: Q.local_zero_density_congruence(2, 4, None, None)
        1/2
        sage: Q.local_zero_density_congruence(3, 6, None, None)
        0
        sage: Q.local_zero_density_congruence(3, 9, None, None)
        2/9

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q.local_zero_density_congruence(2, 2, None, None)
        0
        sage: Q.local_zero_density_congruence(2, 4, None, None)
        1/4
        sage: Q.local_zero_density_congruence(3, 6, None, None)
        0
        sage: Q.local_zero_density_congruence(3, 9, None, None)
        8/81
    """
    #  DIAGNOSTIC
    verbose(" In local_zero_density_congruence with ")
    verbose(" Q is: \n" + str(self))
    verbose(" p = " + str(p))
    verbose(" m = " + str(m))
    verbose(" Zvec = " + str(Zvec))
    verbose(" NZvec = " + str(NZvec))

    #  Put the Zvec congruence condition in a standard form
    if Zvec is None:
        Zvec = []

    n = self.dim()

    #  Sanity Check on Zvec and NZvec:
    #  -------------------------------
    Sn = Set(range(n))
    if (Zvec is not None) and (len(Set(Zvec) + Sn) > n):
        raise RuntimeError("Zvec must be a subset of {0, ..., n-1}.")
    if (NZvec is not None) and (len(Set(NZvec) + Sn) > n):
        raise RuntimeError("NZvec must be a subset of {0, ..., n-1}.")

    p2 = p * p

    #  Check some conditions for no zero-type solutions to exist
    if m % p2 or NZvec is not None:
        return 0

    #  Use the reduction procedure to return the result
    return self.local_density_congruence(p, m / p2, None, None) / p**(self.dim() - 2)


def local_badI_density_congruence(self, p, m, Zvec=None, NZvec=None):
    """
    Find the Bad-type I local density of `Q` representing `m` at `p`.
    (Assuming that `p > 2` and `Q` is given in local diagonal form.)

    INPUT:

    - ``self`` -- quadratic form `Q`, assumed to be block diagonal and `p`-integral

    - ``p`` -- a prime number

    - ``m`` -- integer

    - ``Zvec``, ``NZvec`` -- non-repeating lists of integers in ``range(self.dim())`` or ``None``

    OUTPUT: a rational number

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.local_badI_density_congruence(2, 1, None, None)
        0
        sage: Q.local_badI_density_congruence(2, 2, None, None)
        1
        sage: Q.local_badI_density_congruence(2, 4, None, None)
        0
        sage: Q.local_badI_density_congruence(3, 1, None, None)
        0
        sage: Q.local_badI_density_congruence(3, 6, None, None)
        0
        sage: Q.local_badI_density_congruence(3, 9, None, None)
        0

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q.local_badI_density_congruence(2, 1, None, None)
        0
        sage: Q.local_badI_density_congruence(2, 2, None, None)
        0
        sage: Q.local_badI_density_congruence(2, 4, None, None)
        0
        sage: Q.local_badI_density_congruence(3, 2, None, None)
        0
        sage: Q.local_badI_density_congruence(3, 6, None, None)
        0
        sage: Q.local_badI_density_congruence(3, 9, None, None)
        0

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,3,9])
        sage: Q.local_badI_density_congruence(3, 1, None, None)
        0
        sage: Q.local_badI_density_congruence(3, 3, None, None)
        4/3
        sage: Q.local_badI_density_congruence(3, 6, None, None)
        4/3
        sage: Q.local_badI_density_congruence(3, 9, None, None)
        0
        sage: Q.local_badI_density_congruence(3, 18, None, None)
        0
    """
    #  DIAGNOSTIC
    verbose(" In local_badI_density_congruence with ")
    verbose(" Q is: \n" + str(self))
    verbose(" p = " + str(p))
    verbose(" m = " + str(m))
    verbose(" Zvec = " + str(Zvec))
    verbose(" NZvec = " + str(NZvec))

    #  Put the Zvec congruence condition in a standard form
    if Zvec is None:
        Zvec = []

    n = self.dim()

    #  Sanity Check on Zvec and NZvec:
    #  -------------------------------
    Sn = Set(range(n))
    if (Zvec is not None) and (len(Set(Zvec) + Sn) > n):
        raise RuntimeError("Zvec must be a subset of {0, ..., n-1}.")
    if (NZvec is not None) and (len(Set(NZvec) + Sn) > n):
        raise RuntimeError("NZvec must be a subset of {0, ..., n-1}.")

    #  Define the indexing set S_0, and determine if S_1 is empty:
    #  -----------------------------------------------------------
    S0 = []
    S1_empty_flag = True
    # This is used to check if we should be computing BI solutions at all!
    # (We should really to this earlier, but S1 must be nonzero to proceed.)

    #  Find the valuation of each variable (which will be the same over 2x2 blocks),
    #  remembering those of valuation 0 and if an entry of valuation 1 exists.
    for i in range(n):

        #  Compute the valuation of each index, allowing for off-diagonal terms
        if self[i, i] == 0:
            if i == 0:
                val = valuation(self[i, i + 1], p)    # Look at the term to the right
            else:
                if i == n - 1:
                    val = valuation(self[i - 1, i], p)    # Look at the term above
                else:
                    val = valuation(self[i, i + 1] + self[i - 1, i], p)    # Finds the valuation of the off-diagonal term since only one isn't zero
        else:
            val = valuation(self[i, i], p)

        if val == 0:
            S0 += [i]
        elif val == 1:
            S1_empty_flag = False    # Need to have a non-empty S1 set to proceed with Bad-type I reduction...

    #  Check that S1 is non-empty and p|m to proceed, otherwise return no solutions.
    if S1_empty_flag or m % p:
        return 0

    #  Check some conditions for no bad-type I solutions to exist
    if (NZvec is not None) and (len(Set(S0).intersection(Set(NZvec))) != 0):
        return 0

    #  Check that the form is primitive...     WHY DO WE NEED TO DO THIS?!?
    if not S0:
        print(" Using Q = " + str(self))
        print(" and p = " + str(p))
        raise RuntimeError("the form is not primitive")

    #  DIAGNOSTIC
    verbose(" m = " + str(m) + "   p = " + str(p))
    verbose(" S0 = " + str(S0))
    verbose(" len(S0) = " + str(len(S0)))

    #  Make the form Qnew for the reduction procedure:
    #  -----------------------------------------------
    Qnew = deepcopy(self)    # TODO:  DO THIS WITHOUT A copy()
    for i in range(n):
        if i in S0:
            Qnew[i, i] = p * Qnew[i, i]
            if ((p == 2) and (i < n - 1)):
                Qnew[i, i + 1] = p * Qnew[i, i + 1]
        else:
            Qnew[i, i] = Qnew[i, i] / p
            if ((p == 2) and (i < n - 1)):
                Qnew[i, i + 1] = Qnew[i, i + 1] / p

    #  DIAGNOSTIC
    verbose("\n\n Check of Bad-type I reduction: \n")
    verbose(" Q is " + str(self))
    verbose(" Qnew is " + str(Qnew))
    verbose(" p = " + str(p))
    verbose(" m / p = " + str(m / p))
    verbose(" NZvec " + str(NZvec))

    #  Do the reduction
    Zvec_geq_1 = list(Set([i for i in Zvec if i not in S0]))
    if NZvec is None:
        NZvec_geq_1 = NZvec
    else:
        NZvec_geq_1 = list(Set([i for i in NZvec if i not in S0]))

    return QQ(p**(1 - len(S0))) * Qnew.local_good_density_congruence(p, m / p, Zvec_geq_1, NZvec_geq_1)


def local_badII_density_congruence(self, p, m, Zvec=None, NZvec=None):
    """
    Find the Bad-type II local density of `Q` representing `m` at `p`.
    (Assuming that `p > 2` and `Q` is given in local diagonal form.)

    INPUT:

    - ``self`` -- quadratic form `Q`, assumed to be block diagonal and `p`-integral

    - ``p`` -- a prime number

    - ``m`` -- integer

    - ``Zvec``, ``NZvec`` -- non-repeating lists of integers in ``range(self.dim())`` or ``None``

    OUTPUT: a rational number

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.local_badII_density_congruence(2, 1, None, None)
        0
        sage: Q.local_badII_density_congruence(2, 2, None, None)
        0
        sage: Q.local_badII_density_congruence(2, 4, None, None)
        0
        sage: Q.local_badII_density_congruence(3, 1, None, None)
        0
        sage: Q.local_badII_density_congruence(3, 6, None, None)
        0
        sage: Q.local_badII_density_congruence(3, 9, None, None)
        0
        sage: Q.local_badII_density_congruence(3, 27, None, None)
        0

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,3,9,9])
        sage: Q.local_badII_density_congruence(3, 1, None, None)
        0
        sage: Q.local_badII_density_congruence(3, 3, None, None)
        0
        sage: Q.local_badII_density_congruence(3, 6, None, None)
        0
        sage: Q.local_badII_density_congruence(3, 9, None, None)
        4/27
        sage: Q.local_badII_density_congruence(3, 18, None, None)
        4/9
    """
    #  DIAGNOSTIC
    verbose(" In local_badII_density_congruence with ")
    verbose(" Q is: \n" + str(self))
    verbose(" p = " + str(p))
    verbose(" m = " + str(m))
    verbose(" Zvec = " + str(Zvec))
    verbose(" NZvec = " + str(NZvec))

    #  Put the Zvec congruence condition in a standard form
    if Zvec is None:
        Zvec = []

    n = self.dim()

    #  Sanity Check on Zvec and NZvec:
    #  -------------------------------
    Sn = Set(range(n))
    if (Zvec is not None) and (len(Set(Zvec) + Sn) > n):
        raise RuntimeError("Zvec must be a subset of {0, ..., n-1}.")
    if (NZvec is not None) and (len(Set(NZvec) + Sn) > n):
        raise RuntimeError("NZvec must be a subset of {0, ..., n-1}.")

    #  Define the indexing sets S_i:
    #  -----------------------------
    S0 = []
    S1 = []
    S2plus = []

    for i in range(n):

        #  Compute the valuation of each index, allowing for off-diagonal terms
        if self[i, i] == 0:
            if i == 0:
                val = valuation(self[i, i + 1], p)    # Look at the term to the right
            elif i == n - 1:
                val = valuation(self[i - 1, i], p)    # Look at the term above
            else:
                val = valuation(self[i, i + 1] + self[i - 1, i], p)    # Finds the valuation of the off-diagonal term since only one isn't zero
        else:
            val = valuation(self[i, i], p)

        #  Sort the indices into disjoint sets by their valuation
        if (val == 0):
            S0 += [i]
        elif (val == 1):
            S1 += [i]
        elif (val >= 2):
            S2plus += [i]

    #  Check that S2 is non-empty and p^2 divides m to proceed, otherwise return no solutions.
    p2 = p * p
    if not S2plus or m % p2:
        return 0

    #  Check some conditions for no bad-type II solutions to exist
    if (NZvec is not None) and (len(Set(S2plus).intersection(Set(NZvec))) == 0):
        return 0

    #  Check that the form is primitive...      WHY IS THIS NECESSARY?
    if not S0:
        print(" Using Q = " + str(self))
        print(" and p = " + str(p))
        raise RuntimeError("the form is not primitive")

    #  DIAGNOSTIC
    verbose("\n Entering BII routine ")
    verbose(" S0 is " + str(S0))
    verbose(" S1 is " + str(S1))
    verbose(" S2plus is " + str(S2plus))
    verbose(" m = " + str(m) + "   p = " + str(p))

    #  Make the form Qnew for the reduction procedure:
    #  -----------------------------------------------
    Qnew = deepcopy(self)    # TODO:  DO THIS WITHOUT A copy()
    for i in range(n):
        if i in S2plus:
            Qnew[i, i] = Qnew[i, i] / p2
            if (p == 2) and (i < n - 1):
                Qnew[i, i + 1] = Qnew[i, i + 1] / p2

    #  DIAGNOSTIC
    verbose("\n\n Check of Bad-type II reduction: \n")
    verbose(" Q is " + str(self))
    verbose(" Qnew is " + str(Qnew))

    #  Perform the reduction formula
    Zvec_geq_2 = list(Set([i for i in Zvec if i in S2plus]))
    if NZvec is None:
        NZvec_geq_2 = NZvec
    else:
        NZvec_geq_2 = list(Set([i for i in NZvec if i in S2plus]))

    diff = Qnew.local_density_congruence(p, m / p2, Zvec_geq_2, NZvec_geq_2)
    diff -= Qnew.local_density_congruence(p, m / p2, S2plus, NZvec_geq_2)
    return QQ(p**(len(S2plus) + 2 - n)) * diff


def local_bad_density_congruence(self, p, m, Zvec=None, NZvec=None):
    """
    Find the Bad-type local density of `Q` representing
    `m` at `p`, allowing certain congruence conditions mod `p`.

    INPUT:

    - ``self`` -- quadratic form `Q`, assumed to be block diagonal and `p`-integral

    - ``p`` -- a prime number

    - ``m`` -- integer

    - ``Zvec``, ``NZvec`` -- non-repeating lists of integers in ``range(self.dim())`` or ``None``

    OUTPUT: a rational number

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.local_bad_density_congruence(2, 1, None, None)
        0
        sage: Q.local_bad_density_congruence(2, 2, None, None)
        1
        sage: Q.local_bad_density_congruence(2, 4, None, None)
        0
        sage: Q.local_bad_density_congruence(3, 1, None, None)
        0
        sage: Q.local_bad_density_congruence(3, 6, None, None)
        0
        sage: Q.local_bad_density_congruence(3, 9, None, None)
        0
        sage: Q.local_bad_density_congruence(3, 27, None, None)
        0

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,3,9,9])
        sage: Q.local_bad_density_congruence(3, 1, None, None)
        0
        sage: Q.local_bad_density_congruence(3, 3, None, None)
        4/3
        sage: Q.local_bad_density_congruence(3, 6, None, None)
        4/3
        sage: Q.local_bad_density_congruence(3, 9, None, None)
        4/27
        sage: Q.local_bad_density_congruence(3, 18, None, None)
        4/9
        sage: Q.local_bad_density_congruence(3, 27, None, None)
        8/27
    """
    return self.local_badI_density_congruence(p, m, Zvec, NZvec) + self.local_badII_density_congruence(p, m, Zvec, NZvec)

########################################################
#  local_density and local_density_congruence routines #
########################################################


def local_density_congruence(self, p, m, Zvec=None, NZvec=None):
    """
    Find the local density of `Q` representing `m` at `p`,
    allowing certain congruence conditions mod `p`.

    INPUT:

    - ``self`` -- quadratic form `Q`, assumed to be block diagonal and `p`-integral

    - ``p`` -- a prime number

    - ``m`` -- integer

    - ``Zvec``, ``NZvec`` -- non-repeating lists of integers in ``range(self.dim())`` or ``None``

    OUTPUT: a rational number

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q.local_density_congruence(p=2, m=1, Zvec=None, NZvec=None)
        1
        sage: Q.local_density_congruence(p=3, m=1, Zvec=None, NZvec=None)
        8/9
        sage: Q.local_density_congruence(p=5, m=1, Zvec=None, NZvec=None)
        24/25
        sage: Q.local_density_congruence(p=7, m=1, Zvec=None, NZvec=None)
        48/49
        sage: Q.local_density_congruence(p=11, m=1, Zvec=None, NZvec=None)
        120/121

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.local_density_congruence(2, 1, None, None)
        1
        sage: Q.local_density_congruence(2, 2, None, None)
        1
        sage: Q.local_density_congruence(2, 4, None, None)
        3/2
        sage: Q.local_density_congruence(3, 1, None, None)
        2/3
        sage: Q.local_density_congruence(3, 6, None, None)
        4/3
        sage: Q.local_density_congruence(3, 9, None, None)
        14/9
        sage: Q.local_density_congruence(3, 27, None, None)
        2

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,3,9,9])
        sage: Q.local_density_congruence(3, 1, None, None)
        2
        sage: Q.local_density_congruence(3, 3, None, None)
        4/3
        sage: Q.local_density_congruence(3, 6, None, None)
        4/3
        sage: Q.local_density_congruence(3, 9, None, None)
        2/9
        sage: Q.local_density_congruence(3, 18, None, None)
        4/9
    """
    return self.local_good_density_congruence(p, m, Zvec, NZvec) \
        + self.local_zero_density_congruence(p, m, Zvec, NZvec) \
        + self.local_bad_density_congruence(p, m, Zvec, NZvec)


def local_primitive_density_congruence(self, p, m, Zvec=None, NZvec=None):
    """
    Find the primitive local density of `Q` representing
    `m` at `p`, allowing certain congruence conditions mod `p`.

    .. NOTE::

        The following routine is not used internally, but is included for consistency.

    INPUT:

    - ``self`` -- quadratic form `Q`, assumed to be block diagonal and `p`-integral

    - ``p`` -- a prime number

    - ``m`` -- integer

    - ``Zvec``, ``NZvec`` -- non-repeating lists of integers in ``range(self.dim())`` or ``None``

    OUTPUT: a rational number

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q.local_primitive_density_congruence(p=2, m=1, Zvec=None, NZvec=None)
        1
        sage: Q.local_primitive_density_congruence(p=3, m=1, Zvec=None, NZvec=None)
        8/9
        sage: Q.local_primitive_density_congruence(p=5, m=1, Zvec=None, NZvec=None)
        24/25
        sage: Q.local_primitive_density_congruence(p=7, m=1, Zvec=None, NZvec=None)
        48/49
        sage: Q.local_primitive_density_congruence(p=11, m=1, Zvec=None, NZvec=None)
        120/121

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.local_primitive_density_congruence(2, 1, None, None)
        1
        sage: Q.local_primitive_density_congruence(2, 2, None, None)
        1
        sage: Q.local_primitive_density_congruence(2, 4, None, None)
        1
        sage: Q.local_primitive_density_congruence(3, 1, None, None)
        2/3
        sage: Q.local_primitive_density_congruence(3, 6, None, None)
        4/3
        sage: Q.local_primitive_density_congruence(3, 9, None, None)
        4/3
        sage: Q.local_primitive_density_congruence(3, 27, None, None)
        4/3

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,3,9,9])
        sage: Q.local_primitive_density_congruence(3, 1, None, None)
        2
        sage: Q.local_primitive_density_congruence(3, 3, None, None)
        4/3
        sage: Q.local_primitive_density_congruence(3, 6, None, None)
        4/3
        sage: Q.local_primitive_density_congruence(3, 9, None, None)
        4/27
        sage: Q.local_primitive_density_congruence(3, 18, None, None)
        4/9
        sage: Q.local_primitive_density_congruence(3, 27, None, None)
        8/27
        sage: Q.local_primitive_density_congruence(3, 81, None, None)
        8/27
        sage: Q.local_primitive_density_congruence(3, 243, None, None)
        8/27
    """
    return self.local_good_density_congruence(p, m, Zvec, NZvec) \
        + self.local_bad_density_congruence(p, m, Zvec, NZvec)



# sage_setup: distribution = sagemath-gap
# sage.doctest: needs sage.modules
"""
Base for Classical Matrix Groups with GAP
"""

##############################################################################
#       Copyright (C) 2006 David Joyner and William Stein <wstein@gmail.com>
#                     2013 Volker Braun <vbraun.name@gmail.com>
#                     2017 John Palmieri
#                     2023 Matthias Koeppe
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
##############################################################################

from sage.groups.matrix_gps.matrix_group_gap import MatrixGroup_gap
from sage.groups.matrix_gps.named_group import NamedMatrixGroup_generic


class NamedMatrixGroup_gap(NamedMatrixGroup_generic, MatrixGroup_gap):

    def __init__(self, degree, base_ring, special, sage_name, latex_string,
                 gap_command_string, category=None):
        """
        Base class for "named" matrix groups using LibGAP.

        INPUT:

        - ``degree`` -- integer; the degree (number of rows/columns of
          matrices)

        - ``base_ring`` -- ring; the base ring of the matrices

        - ``special`` -- boolean; whether the matrix group is special,
          that is, elements have determinant one

        - ``latex_string`` -- string; the latex representation

        - ``gap_command_string`` -- string; the GAP command to construct
          the matrix group

        EXAMPLES::

            sage: G = GL(2, GF(3))
            sage: from sage.groups.matrix_gps.named_group_gap import NamedMatrixGroup_gap
            sage: isinstance(G, NamedMatrixGroup_gap)
            True
        """
        from sage.libs.gap.libgap import libgap
        group = libgap.eval(gap_command_string)
        MatrixGroup_gap.__init__(self, degree, base_ring, group,
                                 category=category)
        self._special = special
        self._gap_string = gap_command_string
        self._name_string = sage_name
        self._latex_string = latex_string
